'''
bed2fasta.py - get sequences from bed file
==========================================

:Tags: Genomics Intervals Sequences Conversion BED FASTA



Purpose
-------

This script outputs nucleotide sequences for intervals within
a :term:`bed` formatted file using a corresponding genome file.

Usage
-----

A required input to bed2fasta.py is a cgat indexed genome. To obtain an
idexed human reference genome we would type

Example::
   cat hg19.fasta | index_fasta.py hg19 > hg19.log

This file would then serve as the --genome-file when we wish to extract
sequences from a :term:`bed` formatted file.


For example we could now type::

   cat in.bed | python bed2fasta.py --genome-file hg19 > out.fasta

Where we take a set of genomic intervals (e.g. from a human ChIP-seq experiment)
and output their respective nucleotide sequences.


Type::

   python bed2fasta.py --help

for command line help.

Command line options
--------------------

'''
import sys
import cgatcore.experiment as E
import cgat.Bed as Bed
import cgat.IndexedFasta as IndexedFasta
import cgat.Masker as Masker


def main(argv=None):
    if argv is None:
        argv = sys.argv

    parser = E.ArgumentParser(description=__doc__)

    parser.add_argument("--version", action='version', version="1.0")

    parser.add_argument("-g", "--genome-file", dest="genome_file", type=str,
                        help="filename with genomic sequence to retrieve "
                        "sequences from.")

    parser.add_argument("-m", "--masker", dest="masker", type=str,
                        choices=("dust", "dustmasker", "softmask", "none"),
                        help="apply masker to mask output sequences ")

    parser.add_argument("--output-mode", dest="output_mode", type=str,
                        choices=("intervals", "leftright", "segments"),
                        help="what to output. "
                        "'intervals' generates a single sequence for "
                        "each bed interval. 'leftright' generates two "
                        "sequences, one in each direction, for each bed "
                        "interval. 'segments' can be used to output "
                        "sequence from bed12 files so that sequence only covers "
                        "the segements ")

    parser.add_argument("--min-sequence-length", dest="min_length", type=int,
                        help="require a minimum sequence length ")

    parser.add_argument("--max-sequence-length", dest="max_length", type=int,
                        help="require a maximum sequence length ")

    parser.add_argument(
        "--extend-at", dest="extend_at", type=str,
        choices=("none", "3", "5", "both", "3only", "5only"),
        help="extend at 3', 5' or both or no ends. If 3only or 5only "
        "are set, only the added sequence is returned ")

    parser.add_argument(
        "--extend-by", dest="extend_by", type=int,
        help="extend by # bases")

    parser.add_argument(
        "--use-strand", dest="ignore_strand",
        action="store_false",
        help="use strand information and return reverse complement "
        "on intervals located on the negative strand. ")

    parser.set_defaults(
        genome_file=None,
        masker=None,
        output_mode="intervals",
        min_length=0,
        max_length=0,
        extend_at=None,
        extend_by=100,
        ignore_strand=True,
    )

    (args) = E.start(parser)

    if args.genome_file:
        fasta = IndexedFasta.IndexedFasta(args.genome_file)
        contigs = fasta.getContigSizes()
        fasta.setConverter(IndexedFasta.getConverter("zero-both-open"))

    counter = E.Counter()
    ids, seqs = [], []

    E.info("collecting sequences")
    for bed in Bed.setName(Bed.iterator(args.stdin)):
        counter.input += 1

        lcontig = fasta.getLength(bed.contig)

        if args.ignore_strand:
            strand = "+"
        else:
            strand = bed.strand

        if args.output_mode == "segments" and bed.columns == 12:
            ids.append("%s %s:%i..%i (%s) %s %s" %
                       (bed.name, bed.contig, bed.start, bed.end, strand,
                        bed["blockSizes"], bed["blockStarts"]))
            seg_seqs = [fasta.getSequence(bed.contig, strand, start, end)
                        for start, end in bed.toIntervals()]
            seqs.append("".join(seg_seqs))

        elif (args.output_mode == "intervals" or
              args.output_mode == "segments"):
            ids.append("%s %s:%i..%i (%s)" %
                       (bed.name, bed.contig, bed.start, bed.end, strand))
            seqs.append(
                fasta.getSequence(bed.contig, strand, bed.start, bed.end))

        elif args.output_mode == "leftright":
            l = bed.end - bed.start

            start, end = max(0, bed.start - l), bed.end - l
            ids.append("%s_l %s:%i..%i (%s)" %
                       (bed.name, bed.contig, start, end, strand))
            seqs.append(fasta.getSequence(bed.contig, strand, start, end))

            start, end = bed.start + l, min(lcontig, bed.end + l)
            ids.append("%s_r %s:%i..%i (%s)" %
                       (bed.name, bed.contig, start, end, strand))
            seqs.append(fasta.getSequence(bed.contig, strand, start, end))

    E.info("collected %i sequences" % len(seqs))

    masked = Masker.maskSequences(seqs, args.masker)
    args.stdout.write(
        "\n".join([">%s\n%s" % (x, y) for x, y in zip(ids, masked)]) + "\n")

    E.info("masked %i sequences" % len(seqs))

    counter.output = len(seqs)

    E.info("%s" % counter)

    E.stop()


if __name__ == "__main__":
    sys.exit(main(sys.argv))



'''
cgat_script_template.py - template for cgat scripts
====================================================

:Author:
:Tags: Python

Purpose
-------

.. Overall purpose and function of the script>

Usage
-----

.. Example use case

Example::

   python cgat_script_template.py

Type::

   python cgat_script_template.py --help

for command line help.

Command line options
--------------------

'''

import sys
import cgatcore.experiment as E


def main(argv=None):
    """script main.
    parses command line options in sys.argv, unless *argv* is given.
    """

    if argv is None:
        argv = sys.argv

    # setup command line parser
    parser = E.ArgumentParser(description=__doc__)

    parser.add_argument("-t", "--test", dest="test", type=str,
                        help="supply help")

    # add common options (-h/--help, ...) and parse command line
    (args) = E.start(parser, argv=argv)

    # write footer and output benchmark information.
    E.stop()


if __name__ == "__main__":
    sys.exit(main(sys.argv))



"""compute statistics on a table
================================

This tool outputs summary metrics for a tab-separated table.

The table is read into memory, so there is a limit to the size of
table that can be analyzed.

The tool outputs four columns:

metric
    the name of the metric
count
    number of entities
percent
    percent value of metric
info
    additional information, typically the denominator that the
    percent value is computed from

"""

import sys
import re
import pandas
import cgatcore.experiment as E
import cgatcore.iotools as iotools


def compute_table_summary(table):

    nrows = len(table)
    ncolumns = len(table.columns)
    ncells = nrows * ncolumns
    yield "rows", nrows, nrows, "rows"
    yield "columns", ncolumns, ncolumns, "columns"
    ncells = nrows * ncolumns
    yield "cells", ncells, ncells, "cells"
    yield "rows_with_na", nrows - len(table.dropna(axis=0)), nrows, "rows"
    yield "columns_with_na", ncolumns - len(table.dropna(axis=1).columns), ncolumns, "columns"
    yield "cells_with_na", sum(table.isnull().values.ravel()), ncells, "cells"

    for column in table.columns:
        nna = table[column].dropna()
        yield "column_na", nrows - len(nna), nrows, column
        yield "column_unique", len(nna.unique()), None, column


def main(argv=None):

    parser = E.ArgumentParser(description=__doc__)

    parser.add_argument("--version", action='version', version="1.0")

    parser.add_argument(
        "-d", "--delimiter", dest="delimiter", type=str,
        help="delimiter to separate columns ")

    parser.add_argument(
        "-m", "--method", dest="methods", type=str,
        action="append",
        choices=["row-describe", "column-describe"],
        help="additional methods to apply ")

    parser.set_defaults(
        delimiter="\t",
        methods=[],
    )

    (args) = E.start(parser,
                     argv=argv,
                     add_output_options=True)

    if not args.methods:
        args.methods = ["summary"]

    table = pandas.read_csv(args.stdin, args.delimiter)

    args.stdout.write("metric\tcount\tpercent\tinfo\n")

    for method in args.methods:
        label = re.sub("-", "_", method)
        if method == "summary":
            for category, count, denominator, info in compute_table_summary(table):
                args.stdout.write("\t".join(map(str, (
                    category,
                    count,
                    iotools.pretty_percent(count, denominator, na=""),
                    info))) + "\n")
        elif method == "column-describe":
            df = table.describe().T.stack()
            with E.open_output_file(label) as outf:
                outf.write("label\tcategory\tvalue\n")
                df.to_csv(outf, sep="\t")
        elif method == "row-describe":
            df = table.T.describe().stack()
            with E.open_output_file(label) as outf:
                outf.write("label\tcategory\tvalue\n")
                df.to_csv(outf, sep="\t")

    E.stop()


if __name__ == "__main__":
    sys.exit(main())



'''rnaseq_junction_bams2bam.py - convert mappings against junctions to genomic coordinates
========================================================================================

:Tags: Genomics NGS Genesets

Purpose
-------

This script takes as input a BAM file resulting from reads mapped against
a junction database and outputs a :term:`bam` formatted file in genomic
coordinates.

The contigs should be of the format
<chromosome>|<start>|<exon-end>-<exon-start>|<end>|<splice>|<strand>.

<start> - 0-based coordinate of first base
<exon-end> - 0-based coordinate of last base in exon
<exon-start> - 0-based coordinate of first base in exon
<end> - 0-based coordinate of base after last base

Strand can be either ``fwd`` or ``rev``, though sequences in the database
and coordinates are all on the forward strand.

For example ``chr1|1244933|1244982-1245060|1245110|GTAG|fwd`` translates to the
intron ``chr1:1244983-1245060`` in python coordinates.

The input bam-file is supposed to be sorted by read. Only the best
matches are output for each read, were best is defined both in terms
of number of mismatches and number of colour mismatches.

Usage
-----

Example::

   cat input.bam | python rnaseq_junction_bam2bam.py - --log=log > output.bam

Type::

   python rnaseq_junction_bam2bam.py --help

for command line help.

Command line options
--------------------

'''

import sys
import itertools

import cgatcore.experiment as E
import cgatcore.iotools as iotools
import pysam


def main(argv=None):
    """script main.

    parses command line options in sys.argv, unless *argv* is given.
    """

    if not argv:
        argv = sys.argv

    # setup command line parser
    parser = E.ArgumentParser(description=__doc__)

    parser.add_argument("--version", action='version', version="1.0")

    parser.add_argument("-t", "--template-bam-file", dest="filename_genome_bam", type=str,
                        help="input bam file for header information ")

    parser.add_argument("-s", "--contigs-tsv-file", dest="filename_contigs", type=str,
                        help="filename with contig sizes ")

    parser.add_argument("-o", "--colour", dest="colour_mismatches", action="store_true",
                        help="mismatches will use colour differences (CM tag) ")

    parser.add_argument("-i", "--ignore-mismatches", dest="ignore_mismatches", action="store_true",
                        help="ignore mismatches ")

    parser.add_argument("-c", "--remove-contigs", dest="remove_contigs", type=str,
                        help="','-separated list of contigs to remove ")

    parser.add_argument("-f", "--force-output", dest="force", action="store_true",
                        help="force overwriting of existing files ")

    parser.add_argument("-u", "--unique", dest="unique", action="store_true",
                        help="remove reads not matching uniquely ")

    parser.set_defaults(
        filename_genome_bam=None,
        filename_gtf=None,
        filename_mismapped=None,
        remove_contigs=None,
        force=False,
        unique=False,
        colour_mismatches=False,
        ignore_mismatches=False,
    )

    # add common options (-h/--help, ...) and parse command line
    (args) = E.start(parser, argv=argv)

    genomefile, referencenames, referencelengths = None, None, None

    if args.filename_genome_bam:
        genomefile = pysam.AlignmentFile(args.filename_genome_bam, "rb")
    elif args.filename_contigs:
        contigs = iotools.ReadMap(iotools.open_file(args.filename_contigs))
        data = list(zip(*list(contigs.items())))
        referencenames, referencelengths = data[0], list(map(int, data[1]))
    else:
        raise ValueError(
            "please provide either --template-bam-file or --contigs-tsv-file")

    infile = pysam.AlignmentFile("-", "rb")
    outfile = pysam.AlignmentFile("-", "wb", template=genomefile,
                                  referencenames=referencenames,
                                  referencelengths=referencelengths)

    if args.colour_mismatches:
        tag = "CM"
    else:
        tag = "NM"

    nambiguous = 0
    ninput = 0
    nunmapped = 0
    ncigar = 0
    nfull = 0
    noutput = 0

    contig2tid = dict([(y, x) for x, y in enumerate(outfile.references)])

    for qname, readgroup in itertools.groupby(infile, lambda x: x.qname):
        ninput += 1
        reads = list(readgroup)
        if reads[0].is_unmapped:
            nunmapped += 1
            continue

        # filter for best match
        best = min([x.opt(tag) for x in reads])
        reads = [x for x in reads if x.opt(tag) == best]
        if len(reads) > 1:
            nambiguous += 1
            continue

        read = reads[0]

        # reject complicated matches (indels, etc)
        # to simplify calculations below.
        if len(read.cigar) > 1:
            ncigar += 1
            continue

        # set NH flag to latest count
        t = dict(read.tags)
        t['NH'] = 1
        read.tags = list(t.items())

        sname = infile.getrname(read.tid)

        contig, first_exon_start, middle, last_exon_end, splice, strand = sname.split(
            "|")
        first_exon_end, last_exon_start = middle.split("-")
        first_exon_start, first_exon_end, last_exon_start, last_exon_end = list(map(int, (
            first_exon_start, first_exon_end, last_exon_start, last_exon_end)))
        first_exon_end += 1

        total = first_exon_end - first_exon_start + \
            last_exon_end - last_exon_start
        first_exon_length = first_exon_end - first_exon_start

        match1 = first_exon_length - read.pos
        intron_length = last_exon_start - first_exon_end
        match2 = read.qlen - match1

        # match lies fully in one exon - ignore
        if match1 <= 0 or match2 <= 0:
            nfull += 1
            continue

        # increment pos
        read.pos = first_exon_start + read.pos
        read.tid = contig2tid[contig]
        # 3 = BAM_CREF_SKIP
        read.cigar = [(0, match1), (3, intron_length), (0, match2)]

        outfile.write(read)

        noutput += 1

    outfile.close()
    if genomefile:
        genomefile.close()

    c = E.Counter()
    c.input = ninput
    c.output = noutput
    c.full = nfull
    c.cigar = ncigar
    c.ambiguous = nambiguous
    c.unmapped = nunmapped

    E.info("%s" % str(c))

    # write footer and output benchmark information.
    E.stop()


if __name__ == "__main__":
    sys.exit(main(sys.argv))



"""
bed2graph.py - compute the overlap graph between two bed files
==============================================================

:Tags: Python

Purpose
-------

This script ouputs a list of the names of all overlapping intervals
between two bed files.

Usage
-----

Type::

   python bed2graph.py A.bed.gz B.bed.gz > graph.out

for command line help.

Command line options
--------------------

"""

import sys
import cgatcore.experiment as E
import cgatcore.iotools as iotools
import cgat.Bed as Bed


def main(argv=None):
    """script main.

    parses command line options in sys.argv, unless *argv* is given.
    """

    if not argv:
        argv = sys.argv

    # setup command line parser
    parser = E.ArgumentParser(description=__doc__)

    parser.add_argument("--version", action='version', version="1.0")

    parser.add_argument("-o", "--output-section", dest="output", type=str,
                        choices=("full", "name"),
                        help="output either ``full`` overlapping entries, only the ``name``s.")

    parser.set_defaults(
        output="full",
    )

    # add common options (-h/--help, ...) and parse command line
    (args, unknown) = E.start(parser,
                              argv=argv,
                              unknowns=True)

    if len(unknown) != 2:
        raise ValueError("two arguments required")

    if unknown[0] == "-":
        infile1 = args.stdin
    else:
        infile1 = iotools.open_file(unknown[0], "r")

    infile2 = iotools.open_file(unknown[1], "r")

    idx = Bed.readAndIndex(infile2, with_values=True)

    output = args.output
    outfile = args.stdout

    if output == "name":
        outfile.write("name1\tname2\n")
        outf = lambda x: x.fields[0]
    else:
        outf = str

    for bed in Bed.iterator(infile1):
        try:
            overlaps = idx[bed.contig].find(bed.start, bed.end)
        except (KeyError, IndexError):
            # ignore missing contig and zero length intervals
            continue

        for o in overlaps:
            outfile.write("\t".join((outf(bed), outf(o[2]))) + "\n")

    E.stop()


if __name__ == "__main__":
    sys.exit(main(sys.argv))



"""
cat_tables.py - concatenate tables
==================================

:Tags: Python

Purpose
-------

concatenate tables. Headers of subsequent files are ignored.

Usage
-----

Type::

   python <script_name>.py --help

for command line help.

Command line options
--------------------

"""

import sys
import fileinput

import cgatcore.experiment as E


def main(argv=None):
    """script main.

    parses command line options in sys.argv, unless *argv* is given.
    """

    if not argv:
        argv = sys.argv

    # setup command line parser
    parser = E.ArgumentParser(description=__doc__)

    parser.add_argument("--version", action='version', version="1.0")

    parser.set_defaults(
    )

    # add common options (-h/--help, ...) and parse command line
    (args, unknown) = E.start(parser,
                              argv=argv,
                              unknowns=True)

    if len(unknown) == 0 or (len(unknown) == 1 and unknown[0] == "-"):
        infile = args.stdin
    else:
        infile = fileinput.FileInput(args)

    # do sth
    ninput, nskipped, noutput = 0, 0, 0

    header = False

    for line in infile:
        ninput += 1
        if line.startswith("#"):
            pass
        elif not header:
            header = line
        elif line == header:
            nskipped += 1
            continue

        args.stdout.write(line)
        noutput += 1

    E.info("ninput=%i, noutput=%i, nskipped=%i" % (ninput, noutput, nskipped))

    # write footer and output benchmark information.
    E.stop()

if __name__ == "__main__":
    sys.exit(main(sys.argv))



'''bam2fasta.py - sequences aligned to a particular region
==========================================================

:Tags: Genomics NGS Sequences BAM FASTA Conversion

Purpose
-------

This script takes a :term:`bam` formatted file and outputs all reads
overlapping a certain region.

Example
-------

For example::

   cat in.bam | cgat bam2fasta

This command converts the :term:`bam` formatted file in.bam into

Type::

   cgat bam2fasta --help

for command line help.

Command line options
--------------------

'''

import os
import re
import sys
import collections
import pysam
import numpy
import pandas
import cgatcore.experiment as E
import cgatcore.iotools as iotools


# adapted from here: https://bitbucket.org/brentp/biostuff/src/282b504ac9020fe1449e23f800b20b5bd7d12061/nwalign/pairwise.py?at=default&fileviewer=file-view-default
# replace with alignlib version
# gop/gep would be good to group gaps
def global_align(seqj, seqi, gap=-1, match=1, mismatch=-1, nmatch=0):
    """
    """
    assert "-" not in seqi
    assert "-" not in seqj
    UP, LEFT, DIAG, NONE = range(4)
    max_j = len(seqj)
    max_i = len(seqi)

    score = numpy.zeros((max_i + 1, max_j + 1), dtype='f')
    pointer = numpy.zeros((max_i + 1, max_j + 1), dtype='i')
    pointer[0, 0] = NONE
    score[0, 0] = 0.0
    pointer[0, 1:] = LEFT
    pointer[1:, 0] = UP

    score[0, 1:] = gap * numpy.arange(max_j)
    score[1:, 0] = gap * numpy.arange(max_i)

    for i in range(1, max_i + 1):
        ci = seqi[i - 1]
        for j in range(1, max_j + 1):
            cj = seqj[j - 1]

            if cj == ci:
                diag_score = score[i - 1, j - 1] + match
            elif cj == 'N' or ci == 'N':
                diag_score = score[i - 1, j - 1] + nmatch
            else:
                diag_score = score[i - 1, j - 1] + mismatch
            up_score = score[i - 1, j] + gap
            left_score = score[i, j - 1] + gap
            if diag_score >= up_score:
                if diag_score >= left_score:
                    score[i, j] = diag_score
                    pointer[i, j] = DIAG
                else:
                    score[i, j] = left_score
                    pointer[i, j] = LEFT
            else:
                if up_score > left_score:
                    score[i, j] = up_score
                    pointer[i, j] = UP
                else:
                    score[i, j] = left_score
                    pointer[i, j] = LEFT

    align_j = ""
    align_i = ""
    while True:
        p = pointer[i, j]
        if p == NONE:
            break
        s = score[i, j]
        if p == DIAG:
            align_j += seqj[j - 1]
            align_i += seqi[i - 1]
            i -= 1
            j -= 1
        elif p == LEFT:
            align_j += seqj[j - 1]
            align_i += "-"
            j -= 1
        elif p == UP:
            align_j += "-"
            align_i += seqi[i - 1]
            i -= 1
        else:
            raise Exception('wtf!')

    return align_j[::-1], align_i[::-1]


def get_consensus(sequences, ignore_gaps=False, min_gap_proportion=0):
    """if *ignore_gaps* is False, only report a gap as the consensus if it
    constitutes at least *min_gap_proportion* at sites.
    """
    positions = list(zip(*sequences))
    counts_df = pandas.DataFrame([collections.Counter(x) for x in positions]).fillna(0)
    if "-" in counts_df.columns:
        if ignore_gaps:
            counts_df.drop(["-"], axis=1, inplace=True)
        else:
            # reset all gap-counts where gap is less than min_gap_proportion
            if min_gap_proportion > 0:
                threshold = min_gap_proportion * float(len(sequences))
                counts_df.loc[counts_df["-"] < threshold, ["-"]] = 0

    consensus = "".join(counts_df.idxmax(axis=1)).upper()
    return consensus


def get_anchor_consensus(sequences):
    # filter sequences, removing anything with non-majority length
    lengths = [len(x) for x in sequences if len(x) > 0]
    median_length = numpy.median(lengths)
    s = [x for x in sequences if len(x) == median_length]
    return get_consensus(s)


def iterate_bed(bed_file, merge_intervals):
    if merge_intervals:
        contig, start, end = None, None, None
        for bed in bed_file.fetch(parser=pysam.asBed()):
            if contig != bed.contig:
                if contig is not None:
                    yield contig, start, end
                contig = bed.contig
                start, end = bed.start, bed.end
            end = bed.end
        yield contig, start, end
    else:
        for bed in bed_file.fetch(parser=pysam.asBed()):
            yield bed.contig, bed.start, bed.end


def main(argv=None):
    """script main.

    parses command line options in sys.argv, unless *argv* is given.
    """

    if not argv:
        argv = sys.argv

    # setup command line parser
    parser = E.ArgumentParser(description=__doc__)

    parser.add_argument(
        "-e", "--input-bed-file", dest="input_bed_file", type=str,
        help="input file with intervals. Tab-delimited file of intervals "
        "in bed format to restrict analysis to. ")

    parser.add_argument(
        "-m", "--merge-intervals", dest="merge_intervals", action="store_true",
        help="merge intervals in bed file. Useful if you have a site bed-file "
        )

    parser.add_argument(
        "-f", "--reference-fasta-file", dest="reference_fasta_file",
        help="reference genomic sequence in fasta format. "
        )

    parser.add_argument(
        "-c", "--barcode-fasta-file", dest="barcode_fasta_file",
        help="barcode sequence in fasta format. Variable positions "
        "should be marked by N "
        )

    parser.set_defaults(
        reference_fasta_file=None,
        barcode_fasta_file=None,
        merge_intervals=False,
        input_bed_file=None,
        anchor=5,
    )

    # add common options (-h/--help, ...) and parse command line
    (args) = E.start(parser, argv=argv, add_output_options=True)

    if args.stdin != sys.stdin:
        bamfile = args.stdin.name
    elif args:
        if len(args) > 1:
            raise ValueError("multiple bam files provided in arguments")
        bamfile = args[0]
    else:
        bamfile = "-"

    if args.barcode_fasta_file:
        with pysam.FastxFile(args.barcode_fasta_file) as inf:
            barcode_sequence = next(inf).sequence
    else:
        barcode_sequence = None

    if not os.path.exists(args.reference_fasta_file):
        raise OSError("reference fasta file {} does not exist".format(
            args.reference_fasta_file))

    if not os.path.exists(args.input_bed_file):
        raise OSError("input bed file {} does not exist".format(
            args.input_bed_file))

    bed_in = pysam.TabixFile(args.input_bed_file)
    pysam_in = pysam.AlignmentFile(bamfile)
    anchor = args.anchor

    for region_idx, vals in enumerate(iterate_bed(bed_in, args.merge_intervals)):

        if region_idx > 0:
            raise NotImplementedError("output for multiple regions not yet implemented")

        contig, region_start, region_end = vals
        upstream_anchors, downstream_anchors = [], []
        counter = E.Counter()

        unaligned_fn = E.get_output_file("unaligned_{}.fasta".format(region_idx))
        with iotools.open_file(unaligned_fn, "w") as outf:
            for read in pysam_in.fetch(contig, region_start, region_end):
                counter.overlapping_reads += 1
                try:
                    pairs = read.get_aligned_pairs(with_seq=True)
                except ValueError:
                    counter.no_md_tag += 1
                    continue

                map_ref2read_pos = dict((x[1], x[0]) for x in pairs if x[0] is not None)
                map_ref2ref_base = dict((x[1], x[2]) for x in pairs if x[0] is not None)

                upstream_anchor = "".join(
                    map_ref2ref_base.get(x, "") for x in range(region_start - anchor, region_start))

                downstream_anchor = "".join(
                    map_ref2ref_base.get(x, "") for x in range(region_end, region_end + anchor))

                # check if at least one anchor is aligned
                upstream_matches = sum([x.isupper() for x in upstream_anchor])
                downstream_matches = sum([x.isupper() for x in downstream_anchor])

                if upstream_matches < anchor and downstream_matches < anchor:
                    counter.no_anchor += 1
                    continue
                seq = read.query_alignment_sequence

                # collect full length anchors
                upstream_anchor_start, upstream_anchor_end = region_start - anchor, region_start
                downstream_anchor_start, downstream_anchor_end = region_end, region_end + anchor

                if upstream_anchor_start in map_ref2read_pos and upstream_anchor_end in map_ref2read_pos:
                    upstream_anchors.append(
                        seq[map_ref2read_pos[upstream_anchor_start]:map_ref2read_pos[upstream_anchor_end]])
                if downstream_anchor_start in map_ref2read_pos and downstream_anchor_end in map_ref2read_pos:
                    downstream_anchors.append(
                        seq[map_ref2read_pos[downstream_anchor_start]:map_ref2read_pos[downstream_anchor_end]])

                # get region to align
                read_start = min((map_ref2read_pos.get(x, len(seq))
                                  for x in range(region_start - anchor, region_start)))
                if read_start == len(seq):
                    read_start = 0
                read_end = max((map_ref2read_pos.get(x, 0) + 1
                                for x in range(region_end, region_end + anchor)))
                if read_end == 1:
                    read_end = len(seq)
                counter.collected_reads += 1
                outf.write(">{}/{}-{}\n{}\n".format(read.query_name,
                                                    read_start, read_end,
                                                    seq[read_start:read_end]))
        counter.downstream_anchors = len(downstream_anchors)
        counter.upstream_anchors = len(upstream_anchors)

        E.info(counter)

        if counter.overlapping_reads == 0:
            E.warn("no sequences overlapping region")
            continue

        if counter.downstream_anchors == 0 or counter.upstream_anchors == 0:
            E.warn("at least one anchor undefined")
            continue

        if counter.collected_reads == 1:
            E.warn("only single sequence, multiple aligment skipped")
            with iotools.open_file(unaligned_fn) as inf:
                stdout = inf.read()
        else:
            # G-INS-i -> global alignment algorithm
            E.info("starting mafft multiple alignment")
            stdout = E.run("mafft --globalpair --maxiterate 100 --quiet --op 2 --ep 0.5 {}".format(unaligned_fn),
                           return_stdout=True)

        aligned_fn = E.get_output_file("aligned_{}.fasta".format(region_idx))
        with iotools.open_file(aligned_fn, "w") as outf:
            outf.write(stdout)

        mali = stdout.splitlines()
        identifiers = [mali[x] for x in range(0, len(mali), 2)]
        sequences = [mali[x].upper() for x in range(1, len(mali), 2)]
        consensus = get_consensus(sequences)

        E.info("after alignment: consensus={}".format(consensus))

        # gap filtering -> remove highly gappy columns
        consensus = get_consensus(sequences, min_gap_proportion=0.9)

        E.info("after anchor trimming: consensus={}".format(consensus))

        take = [idx for idx, x in enumerate(consensus) if x != "-"]
        sequences = ["".join([s[x] for x in take]) for s in sequences]
        consensus = get_consensus(sequences, min_gap_proportion=0.9)

        E.info("after gap filtering: consensus={}".format(consensus))

        # get anchor consensus and chop it off
        consensus = get_consensus(sequences, ignore_gaps=True)
        upstream_anchor = get_anchor_consensus(upstream_anchors)
        downstream_anchor = get_anchor_consensus(downstream_anchors)

        upstream_anchor_start = consensus.find(upstream_anchor)
        downstream_anchor_start = consensus.rfind(downstream_anchor)

        E.info("anchor consensus (no gaps)={}, upstream={}, downstream={}, upstream_idx={}, downstream_idx={}".format(
            consensus, upstream_anchor, downstream_anchor, upstream_anchor_start, downstream_anchor_start))

        if upstream_anchor_start < 0 or downstream_anchor_start < 0:
            E.warn("can't locate anchor, no output produced")
            continue

        upstream_anchor_end = upstream_anchor_start + len(upstream_anchor)
        if upstream_anchor_end >= downstream_anchor_start:
            E.warn("anchor not in correct order, no output produced")
            continue

        sequences = [x[upstream_anchor_end:downstream_anchor_start] for x in sequences]
        consensus = get_consensus(sequences)

        E.info("after anchor trimming: consensus={}".format(consensus))

        truncated_fn = E.get_output_file("aligned_truncated_{}.fasta".format(region_idx))
        with iotools.open_file(truncated_fn, "w") as outf:
            outf.write("\n".join(
                "{}\n{}\n".format(x, y) for x, y in zip(identifiers, sequences)))

        positions = list(zip(*sequences))
        bases = ["A", "C", "G", "T"]
        df = pandas.DataFrame([collections.Counter(x) for x in positions]).fillna(0)
        for missing_base in [x for x in bases if x not in df.columns]:
            df[missing_base] = 0
        df["gapped_depth"] = df.sum(axis=1)
        df["depth"] = df[bases].sum(axis=1)
        df["consensus"] = df[bases].idxmax(axis=1)
        df["consensus_counts"] = df.lookup(df.index, df.consensus)
        df["consensus_support"] = df.consensus_counts / df.depth
        df["offconsensus_counts"] = df.depth - df.consensus_counts
        df.loc[df.consensus_counts == 0, "consensus"] = "N"
        df["region_id"] = region_idx

        # replace "gap" consensus positions with + character
        alignment = global_align(re.sub("-", "+", consensus), barcode_sequence)
        E.info("alignment: consensus {}".format(alignment[0]))
        E.info("alignment: barcode   {}".format(alignment[1]))

        barcode_idx = 0
        deleted_barcode_bases = []
        rows = []
        for c, b in zip(*alignment):
            if c == "-":
                deleted_barcode_bases.append(barcode_idx)
                barcode_idx += 1
            elif b == "N":
                rows.append((barcode_idx, "variable"))
                barcode_idx += 1
            elif b == "-":
                rows.append(("", "insertion"))
            elif b == c:
                rows.append((barcode_idx, "fixed-match"))
                barcode_idx += 1
            else:
                rows.append((barcode_idx, "fixed-mismatch"))
                barcode_idx += 1

        alignment_df = pandas.DataFrame.from_records(rows,
                                                     columns=["barcode_pos", "barcode_class"])

        assert len(alignment_df) == len(df)
        df = pandas.concat([df, alignment_df], axis=1)
        with E.open_output_file("pileup") as outf:
            df.to_csv(outf, sep="\t", index=True, index_label="position")

        observed_barcode_sequence = "".join(df[df.barcode_class == "variable"].consensus)
        headers = df.consensus_support.describe().index
        eval_df = df.loc[df.barcode_class.isin(("variable", "fixed-match", "fixed-mismatch")), ]
        median_consensus_depth = eval_df.consensus_counts.median()
        # zero stuff out if depth is low
        if median_consensus_depth <= 2:
            deleted_barcode_bases = []

        outf = args.stdout
        # modules to recover partial bar-codes
        outf.write("\t".join(map(str,
                                 ["barcode", "ndeleted_barcode_bases", "deleted_barcode_bases"] +
                                 ["support_{}".format(x) for x in headers] +
                                 ["counts_{}".format(x) for x in headers] +
                                 ["offcounts_{}".format(x) for x in headers])) + "\n")

        outf.write("\t".join(map(str, [
            observed_barcode_sequence,
            len(deleted_barcode_bases),
            ",".join(map(str, deleted_barcode_bases))] +
                                 eval_df.consensus_support.describe().tolist() +
                                 eval_df.consensus_counts.describe().tolist() +
                                 eval_df.offconsensus_counts.describe().tolist())) + "\n")

    E.stop()


if __name__ == "__main__":
    sys.exit(main(sys.argv))



'''fasta2fastq.py - simulate reads from fasta
=====================================

:Tags: Sequences

Purpose
-------

Simulate illumina sequence reads from a fasta file. The number of
reads per entry is randomly selected from the range given. The primary
use case is expected to be the generation of simulation RNA-Seq reads

For RNA-Seq simulations, the premrna-fraction option allows the user to specify
what fraction of the transcripts originate from pre-mRNA. The user must also
supply a second fasta in the same order for the pre-mRNA
(--infile-premrna-fasta). The simulation assumes all pre-mRNA are full length
which is not likely to be the case for real RNA-Seq.
Note: This may lead to many more reads which align to the mRNA than the
apparent ground truth count. It is therefore best to keep the pre-mRNA fraction
low (recommend 0.01).

--counts-method determine whether the --min-counts and --max--counts methods
relate to the number of reads or copies per entry. If copies, then the number
of reads/read pairs per transcript will be determined by the copy number, the
length of the transcript and the fragment length. E.g for copy number = 2,
transcript length = 1000 bp, read length = 100 bp, paired end, 300bp insert
size, the number of reads will be 2 * (1000 / (( 2 * 100) + 300)) = 4

Options
-------

--output-paired-end
   generate paired-end reads (defaults to single end)

--counts-method
  simulate a ground truth number of reads or copies per transcript
  if reads, then min and max specify the number of reads per entry
  if copies, then min and specify the sequencing depth per entry

--counts-min
   the minimum number of reads to simulate for each fasta entry

--counts-max
   the maximum number of reads to simulate for each fasta entry

--sequence-error-phred
   the sequencing error rate (phred scale)

--output-read-length
   the length of the outputted reads

--output-counts
   filename for counts per fasta entry

--output-quality-format
   the format of the sequence qualities (+33 = Sanger)

--insert-length-mean
   the mean insert length

--insert-length-sd
   the standard deviation for the insert length

--premrna-fraction
   the fraction of reads to simulate from pre-mRNA. Default is 0.
   If set, must provide a pre-mRNA fasta file with:
      --infile-premrna-fasta

If generating paired end reads, the second outfile must be specified with:

--output-fastq2


Usage
-----

Recommend sending logging to separate outfile to keep fastq outfile
clean of comments (see example below)

Example::

   cat transcripts.fa | python fasta2fastq.py
   --output-counts=simulation_counts.tsv -L simulation.log
   > simulation_reads.fastq

Type::

   python fasta2fastq.py --help

for command line help.


Important note for generating reads for simulations
---------------------------------------------------
Currently, the output is non-random, e.g it's in the order of the
fasta input. If you want the fastq to be random pipe the output to
sort -R like so:

   cat transcripts.fa | python fasta2fastq.py
   --output-counts=simulation_counts.tsv -L simulation.log |
   paste - - - - |sort -R | sed 's/\t/\n/g' > simulation_reads_random.fastq

If you're outputting paired end fastqs, you can use the following
command to randomise the order by keep the fastq entris paired:

    paste <(zcat %(fastq1_ordered)s) <(zcat %(fastq2_ordered)s) |
    paste - - - - | sort -R | awk -F'\t' '{OFS="\n"; print $1,$3,$5,$7 >
    "%(fastq1_random)s"; print $2,$4,$6,$8 > "%(fastq2_random)s"}'

'''
import sys
import random
import numpy as np
import collections

import cgatcore.experiment as E
import cgatcore.iotools as iotools

import cgat.FastaIterator as FastaIterator


def addSeqErrors(read=None, error_rate=10):
    ''' add sequencing errors to a read.
    Error rates are Phred scaled, so 30 = 1/1000'''

    error_rate = 10**(error_rate/-10.0)

    errors_dict = {"G": ["C", "T", "A"],
                   "C": ["G", "T", "A"],
                   "T": ["C", "G", "A"],
                   "A": ["C", "T", "G"],
                   "N": ["C", "T", "G", "A"]}

    probs = np.random.rand(len(read))
    return "".join([base if prob > error_rate and base != "N"
                    else random.choice(errors_dict[base])
                    for prob, base in zip(probs, read)])


def reverseComp(seq):
    ''' return the reverse complement sequence '''

    comp = {"G": "C",
            "C": "G",
            "A": "T",
            "T": "A",
            "N": "N"}

    return "".join([comp[base] for base in seq[::-1]])


def generateRead(entry, read_length=50, error_rate=40, paired=False,
                 insert_mean=0, insert_sd=1):
    ''' generate a read (or read pair) at random from a fasta entry for
    the given read length with sequencing errors according to error
    rate'''

    if paired:

        position = "not_OK"

        while position != "OK":

            r1_start = random.randint(0, len(entry)-read_length)
            r2_start = (r1_start + read_length +
                        int(np.random.normal(insert_mean, insert_sd)))

            if (r2_start <= (len(entry) - read_length) and r2_start >= r1_start):

                position = "OK"

                read1 = entry[r1_start: r1_start+read_length]
                read2 = reverseComp(
                    entry[r2_start: r2_start+read_length])

                final_read1 = addSeqErrors(read1, error_rate)
                final_read2 = addSeqErrors(read2, error_rate)

                return final_read1, final_read2

    else:
        start = random.randint(0, len(entry)-read_length)
        read = entry[start:start+read_length]

        final_read = addSeqErrors(read, error_rate)

        return final_read


def getTitle(entry):
    ''' return the title for an entry'''
    return entry.title.split()[0]


def main(argv=None):
    if argv is None:
        argv = sys.argv

    parser = E.ArgumentParser(description=__doc__)

    parser.add_argument(
        "--output-quality-format", dest="q_format", type=int,
        help="sequence quality format, e.g 33 = +33/Sanger")

    parser.add_argument(
        "--output-paired-end", dest="paired", action="store_true",
        help="generate paired end reads")

    parser.add_argument(
        "--insert-length-mean", dest="insert_mean", type=float,
        help="mean insert length.")

    parser.add_argument(
        "--insert-length-sd", dest="insert_sd", type=float,
        help="insert length standard deviation.")

    parser.add_argument(
        "--counts-method", dest="counts_method", type=str,
        choices=("reads", "copies"),
        help="simulate a ground truth number of reads per entry or"
        "copies per entry.")

    parser.add_argument(
        "--counts-min", dest="counts_min", type=float,
        help="minimum number of reads/read pairs per fasta entry"
        "or copies per entry.")

    parser.add_argument(
        "--counts-max", dest="counts_max", type=float,
        help="maximum number of reads/read pairs per fasta entry "
        "or copies per entry.")

    parser.add_argument(
        "--output-read-length", dest="read_length", type=int,
        help="read length.")

    parser.add_argument(
        "--sequence-error-phred", dest="phred", type=int,
        help="phred quality score.")

    parser.add_argument(
        "--output-counts", dest="output_counts", type=str,
        help="name for counts outfile.")

    parser.add_argument(
        "--output-fastq2", dest="fastq2_out", type=str,
        help="filename for second fastq outfile.")

    parser.add_argument(
        "--premrna-fraction", dest="premrna_fraction", type=float,
        help="the fraction of reads to simulate from pre-mRNA")

    parser.add_argument(
        "--infile-premrna-fasta", dest="premrna_fasta", type=str,
        help="filename for pre-mRNA fasta.")

    parser.set_defaults(
        q_format=33,
        paired=False,
        insert_mean=0,
        insert_sd=1,
        counts_method="reads",
        counts_min=1,
        counts_max=1,
        read_length=50,
        fastq2_out=None,
        output_counts=None,
        phred=30,
        premrna_fraction=0,
        premrna_fasta=None
    )

    (args) = E.start(parser)

    if args.paired:
        assert args.fastq2_out, ("must specify a second fastq outfile for "
                                 "paired end (--output-fastq2)")
        outf2 = iotools.open_file(args.fastq2_out, "w")

    if args.premrna_fraction:
        assert args.premrna_fasta, ("must specfify the location of the"
                                    "fasta file for the pre-mRNA")

    # the sequence quality string will always be the same so define here
    sequence_quality = chr(args.q_format + args.phred)
    qual = "".join([sequence_quality] * args.read_length)

    if args.premrna_fraction:
        iterator = FastaIterator.iterate_together(
            args.stdin, iotools.open_file(args.premrna_fasta))
    else:
        iterator = FastaIterator.FastaIterator(args.stdin)

    # set a cut off of twice the read/pair length for short entries
    if args.paired:
        minimum_entry_length = (
            2 * ((args.read_length * 2) + args.insert_mean))
    else:
        minimum_entry_length = 2 * args.read_length

    c = collections.Counter()
    counts = collections.Counter()
    copies = collections.Counter()

    for f_entry in iterator:

        if args.premrna_fraction:

            assert getTitle(f_entry[0]) == getTitle(f_entry[1]), (
                "entry ids do not match: %s != %s" % (
                    f_entry[0].title, f_entry[1].title))
            entry = f_entry[0]
            pre_entry = f_entry[1]

        else:
            entry = f_entry

        # reject short fasta entries
        if len(entry.sequence) < minimum_entry_length:
            E.info("skipping short transcript: %s length=%i"
                   % (entry.title, len(entry.sequence)))
            c['skipped'] += 1
            continue

        else:
            c['not_skipped'] += 1

        if args.paired:
            fragment_length = (
                (2 * args.read_length) + args.insert_mean)
        else:
            fragment_length = args.read_length

        reads_per_entry = float(len(entry.sequence)) / fragment_length

        if args.counts_method == "reads":
            n_reads = random.randint(args.counts_min,
                                     args.counts_max + 1)

            n_copies = float(n_reads) / reads_per_entry

            if args.premrna_fraction:
                n_reads_pre = int(round(n_reads * args.premrna_fraction))

        elif args.counts_method == "copies":

            # random float [0-1]
            rand = np.random.random_sample()
            n_copies = (args.counts_min +
                        (rand * (args.counts_max - args.counts_min)))

            n_reads = int(round(n_copies * reads_per_entry, 0))

            # as n_reads must be rounded to int, need to redefine n_copies
            n_copies = float(n_reads) / reads_per_entry

            if args.premrna_fraction:
                reads_per_pre_entry = (float(len(pre_entry.sequence)) /
                                       fragment_length)
                n_copies_pre = n_copies * args.premrna_fraction
                n_reads_pre = int(round(n_copies_pre * reads_per_pre_entry, 0))
                # as n_reads_pre must be rounded to int, need to
                # redefine n_copies_pre
                n_copies_pre = float(n_reads_pre) / reads_per_pre_entry

        entry_id = getTitle(entry)

        counts[entry_id] = n_reads
        copies[entry_id] = n_copies

        if "N" in entry.sequence.upper():
            E.warn("fasta entry %s contains unknown bases ('N')" % entry_id)

        for i in range(0, n_reads):

            read = generateRead(entry=entry.sequence.upper(),
                                read_length=args.read_length,
                                error_rate=args.phred,
                                paired=args.paired,
                                insert_mean=args.insert_mean,
                                insert_sd=args.insert_sd)

            if args.paired:
                r1, r2 = read
                h1 = "@%s_%i/1" % (entry_id, i)
                h2 = "@%s_%i/2" % (entry_id, i)

                args.stdout.write("\n".join((h1, r1, "+", qual)) + "\n")
                outf2.write("\n".join((h2, r2, "+", qual)) + "\n")

            else:
                h = "@%s_%i/1" % (entry_id, i)

                args.stdout.write("\n".join((h, read, "+", qual)) + "\n")

        if args.premrna_fraction:
            c['pre_counts'] += n_reads_pre
            c['pre_copies'] += n_copies_pre

            for i in range(0, n_reads_pre):

                read = generateRead(entry=pre_entry.sequence.upper(),
                                    read_length=args.read_length,
                                    error_rate=args.phred,
                                    paired=args.paired,
                                    insert_mean=args.insert_mean,
                                    insert_sd=args.insert_sd)

                if args.paired:
                    r1, r2 = read
                    h1 = "@%s_pre-mRNA_%i/1" % (entry_id, i)
                    h2 = "@%s_pre-mRNA_%i/2" % (entry_id, i)

                    args.stdout.write("\n".join((h1, r1, "+", qual)) + "\n")
                    outf2.write("\n".join((h2, r2, "+", qual)) + "\n")

                else:
                    h = "@%s_pre-mRNA_%i/1" % (entry_id, i)

                    args.stdout.write("\n".join((h, read, "+", qual)) + "\n")

    if args.paired:
        outf2.close()

    with iotools.open_file(args.output_counts, "w") as counts_out:

        counts_out.write("%s\n" % "\t".join(("id", "read_count", "tpm")))

        sum_copies = sum(copies.values())
        sum_counts = sum(counts.values())

        for entry_id, count in counts.items():
            tpm = 1000000 * (float(copies[entry_id]) / sum_copies)
            counts_out.write(
                "%s\n" % "\t".join(map(str, (entry_id, count, tpm))))

    E.info("Reads simulated for %i fasta entries, %i entries skipped"
           % (c['not_skipped'], c['skipped']))

    E.info("Simulated: %i reads (%i mRNA, %i pre-mRNA), "
           "%f transcripts (%f mRNA, %f pre-mRNA)" % (
               sum_counts + c['pre_counts'], sum_counts, c['pre_counts'],
               sum_copies + c['pre_copies'], sum_copies, c['pre_copies']))

    E.stop()


if __name__ == "__main__":
    sys.exit(main(sys.argv))



'''
bed2annotator.py - convert bed to annotator format
==================================================

:Tags: Python

Purpose
-------

This script converts a bed file into annotator compatible regions. Depending on the option --section
this script will create:

   segments
      a segments file

   annotations
      a file with annotations. Each bed track is a separate annotation.

   workspace
      a file with a workspace

Usage
-----

Example::

   python bed2annotator2tsv.py --help

Type::

   python bed2annotator2tsv.py --help

for command line help.

Command line options
--------------------

'''
import sys
import re
import itertools
import collections

import cgatcore.experiment as E
import cgat.Bed as Bed
import cgat.IndexedFasta as IndexedFasta


def main(argv=None):
    """script main.

    parses command line options in sys.argv, unless *argv* is given.
    """

    if argv is None:
        argv = sys.argv

    parser = E.ArgumentParser(description=__doc__)

    parser.add_argument("-g", "--genome-file", dest="genome_file", type=str,
                        help="filename with genome.")

    parser.add_argument("-f", "--features", dest="features", type=str,
                        help="feature to collect.")

    parser.add_argument("-i", "--files", dest="files", action="append",
                        help="use multiple annotations.")

    parser.add_argument("-a", "--annotations", dest="annotations", type=str,
                        help="aggregate name for annotations if only single file is provided from STDIN.")

    parser.add_argument("--map-tsv-file", dest="input_filename_map", type=str,
                        help="filename with a map of gene_ids to categories.")

    parser.add_argument("-l", "--max-length", dest="max_length", type=str,
                        help="maximum segment length.")

    parser.add_argument("-m", "--merge-overlapping", dest="merge", action="store_true",
                        help="merge overlapping bed segments.")

    parser.add_argument("-s", "--section", dest="section", type=str,
                        choices=("segments", "annotations", "workspace"),
                        help="annotator section.")

    parser.add_argument("--subset", dest="subsets", type=str, action="append",
                        help="add filenames to delimit subsets within the gff files. The syntax is filename.gff,label,filename.ids.")

    parser.set_defaults(
        genome_file=None,
        feature=None,
        remove_random=True,
        section="segments",
        annotations="annotations",
        max_length=100000,
        files=[],
        subsets=[],
        input_filename_map=None,
        merge=False,
    )

    (args, unknown) = E.start(parser,
                              unknowns=True)

    args.files += unknown
    if len(args.files) == 0:
        args.files.append("-")
    args.files = list(
        itertools.chain(*[re.split("[,; ]+", x) for x in args.files]))

    if args.subsets:
        subsets = collections.defaultdict(list)
        for s in args.subsets:
            filename_gff, label, filename_ids = s.split(",")
            subsets[filename_gff].append((label, filename_ids))
        args.subsets = subsets

    if args.genome_file:
        fasta = IndexedFasta.IndexedFasta(args.genome_file)
    else:
        fasta = None

    if args.section == "segments":
        prefix = "##Segs"
    elif args.section == "annotations":
        prefix = "##Id"
    elif args.section == "workspace":
        prefix = "##Work"
    else:
        raise ValueError("unknown section %s" % args.section)

    if args.max_length:
        max_length = args.max_length
    else:
        max_length = 0

    ninput, ntracks, ncontigs, nsegments, ndiscarded = 0, 0, 0, 0, 0

    if args.section in ("annotations"):
        contigs = set()
        it = itertools.groupby(
            Bed.iterator(args.stdin), key=lambda x: x.track["name"])

        map_track2segments = {}
        for track, beds in it:
            ntracks += 1
            map_track2segments[track] = []
            first_segment = nsegments

            beds = list(beds)

            if args.merge:
                beds = Bed.merge(beds)

            for bed in beds:
                contig, start, end = bed.contig, bed.start, bed.end

                if args.remove_random and "random" in contig:
                    continue

                if max_length > 0 and end - start > max_length:
                    ndiscarded += 1
                    continue

                contigs.add(contig)
                map_track2segments[track].append(nsegments)
                args.stdout.write(
                    "%s\t%i\t%s\t(%i,%i)\n" % (prefix, nsegments, contig, start, end))
                nsegments += 1

            args.stdout.write("##Ann\t%s\t%s\n" % (
                track, "\t".join(["%i" % x for x in range(first_segment, nsegments)])))
            E.info("track %s: annotated with %i segments" %
                   (track, nsegments - first_segment))

        ncontigs = len(contigs)
        E.info("ninput=%i, ntracks=%i, ncontigs=%i, nsegments=%i, ndiscarded=%i" %
               (ninput, ntracks, ncontigs, nsegments, ndiscarded))

    E.stop()


if __name__ == "__main__":
    sys.exit(main(sys.argv))



"""=================================================================
genome_bed.py - Create a bed file tiling a genome from a fai file
=================================================================

:Tags: Python

This program takes an indexed genome and creates windows of a certain
size.

It also takes two input parameters: the window/tile size (bases) and
the shift size.  By default the shift size is equal to the window
size.  The default window size is 1000.

Usage
=====

   python genome_bed -g <genome.fai> -o <output.bed> -w window size -s shift size


Command line options
====================

"""


import cgatcore.experiment as E
import sys
import cgat.IndexedFasta as IndexedFasta


def main(argv=None):
    """script main.

    parses command line options in sys.argv, unless *argv* is given.
    """

    if not argv:
        argv = sys.argv

    # setup command line parser
    parser = E.ArgumentParser(description=__doc__)

    parser.add_argument("--version", action='version', version="1.0")

    parser.add_argument(
        "-g", "--genome-file", dest="genome_file", type=str,
        help="filename with Samtools indexed genome.")
    parser.add_argument(
        "-w", "--window-size", dest="window", type=int,
        help="Window size for tiling.")
    parser.add_argument(
        "-s", "--shift-size", dest="shift", type=int,
        help="Window shift for tiling.")

    parser.set_defaults(
        genome_file=None,
        window=1000,
        shift=1000,
        output_file=None,
    )

    # add common options (-h/--help, ...) and parse command line
    (args) = E.start(parser, argv=argv)

    ninput, nunchanged, nchanged = 0, 0, 0

    # Open input file
    E.info("Opening input file: %s" % args.genome_file)
    fasta = IndexedFasta.IndexedFasta(args.genome_file)
    contigs = fasta.getContigSizes(with_synonyms=False)

    # Open output file
    bed = args.stdout
    shift = args.shift

    # Loop over input files and convert to soft clipped
    nwindows = 0
    ncontigs = 0
    for contig, stop in contigs.items():
        ncontigs += 1
        i = 0
        while (i < stop):
            j = min(i + args.window, stop)
            #bed.write( """%(contig)s\t%(i)i\t%(j)i\t%(contig)s:%(i)i..%(j)i\n""" % locals() )
            bed.write( """%(contig)s\t%(i)i\t%(j)i\n""" % locals() )
            nwindows += 1
            i += shift

    # Report statistics
    E.info("ncontigs=%i, nwindows=%i" % (ncontigs, nwindows))

    # write footer and output benchmark information.
    E.stop()

if __name__ == "__main__":
    sys.exit(main(sys.argv))



'''
extract_stats.py - extract and process tables from csvDB
========================================================

Purpose
-------

Extract tables from sqlite databases and process

Usage
-----

.. Example use case

Example::

   python extract_stats.py

Type::

   python extract_stats.py --help

for command line help.

Command line options
--------------------

tasks
+++++

`extract_table` - extract a table from a database
                  containing relevant information

`get_coverage` - calculate gene/transcript model
                 coverage stats - only works on single cell data
                 with filename format <seqRun>_<plate>_<well>_<mappper>

`aggregate` - aggregate together multiple stats tables,
              and select relevant measures

.. todo::
    Needs to be refactored to use sqlalchemy or Database instead of
    explicitely wrapping database frontends.


'''

import sys
import pandas
import numpy
import re
import cgatcore.experiment as E
import cgatcore.database as database


def getTableFromDb(database_url, table):
    '''
    Get a table from a database with pandas
    '''

    dbhandle = database.connect(url=database_url)
    df = pandas.read_sql("SELECT * FROM {}".format(table), con=dbhandle)
    df.index = df["track"]
    df.drop(labels="track", inplace=True, axis=1)

    return df


def cleanStatsTable(stats_file):
    '''
    Take in a table containing aggregated stats
    and clean by removing duplicate columns
    '''
    # , mangle_dupe_cols=False)
    # AH: disabled, because "ValueError: Setting mangle_dupe_cols=False is not supported yet"
    df = pandas.read_table(stats_file, sep="\t", header=0,
                           index_col=None)

    # drop duplicates is case sensitive, convert all to
    # same case - SQL is not case sensitive so will throw
    # a hissy fit for same column names in different cases
    df.columns = [cx.lower() for cx in df.columns]
    df = df.T.drop_duplicates().T
    df.index = df["track"]
    return df


def extractTranscriptCounts(con, table):
    '''
    Extract transcript model counts for a
    given sample

    Arguments
    ---------
    con: sqlite.connection
      An SQLite connection

    table: string
      the table to extract the transcript counts
      from.

    Returns
    -------
    coverages: pandas.Core.Series
    '''

    statement = '''
    SELECT coverage_sense_pcovered
    FROM %(table)s
    WHERE coverage_sense_nval > 0;
    ''' % locals()

    coverages = pandas.read_sql(statement, con)
    coverages = coverages.loc[:, "coverage_sense_pcovered"]
    return coverages


def summariseOverBins(coverages, bins):
    '''
    Summarise model coverages over a set of bins

    Argumnets
    ---------
    coverages: pandas.Core.Series
      coverages over gene/transcripts

    bins: list
      values corresponding to percentage bins

    Returns
    -------
    freqs: numpy.array
      frequency array of coverages over percentiles
    '''

    freqs = numpy.zeros(shape=len(bins), dtype=numpy.float64)
    for i in range(len(bins)):
        if i == 0:
            hits = coverages <= bins[i]
        else:
            hits = (coverages <= bins[i]) & (coverages > bins[i-1])

        freqs[i] = len(coverages[hits])

    return freqs


def getModelCoverage(database_url, table_regex, model_type="transcript"):
    '''
    Compute transcript model coverage stats

    Arguments
    ---------
    database_url: string
      database containing transcript counts

    table_regex: string
      regular expression for transcript count table

    model_type: string
      calculate coverages over either transcripts or
      genes.  Default is gene models

    Returns
    -------
    coverage_df: Pandas.Core.DataFrame
      model coverage stats summarised for each cell
    '''

    # need to regex for all the tables, one for each sample
    # fetch_all returns a list of tuples
    dbhandle = Database.connect(database_url)
    cc = dbhandle.execute("SELECT name FROM sqlite_master WHERE type='table';")

    tab_reg = re.compile(table_regex)
    table_list = [tx[0] for tx in cc.fetchall() if re.search(tab_reg, tx[0])]

    # pull out counts for each cell and compute coverages
    bins = range(0, 101)
    cov_dict = {}
    for tab in table_list:
        covs = extractTranscriptCounts(dbhandle, tab)
        freq_array = summariseOverBins(covs, bins)
        cov_dict[tab] = freq_array

    coverage_df = pandas.DataFrame(cov_dict).T
    # create a regex group to remove superfluous characters
    # from the track names
    ix_re = re.compile("_(?P<run>\d+)_(?P<plate>\d+)_(?P<well>\d+)_(?P<mapper>\S+)_transcript_counts")
    re_matches = [re.match(ix_re, ix) for ix in coverage_df.index]
    indx = ["%s_%s-%s.%s" % rm.group(1, 2, 3, 4) for rm in re_matches]
    coverage_df.index = indx
    coverage_df.columns = ["Bin%i" % bx for bx in coverage_df.columns]
    return coverage_df


def main(argv=None):
    """script main.
    parses command line options in sys.argv, unless *argv* is given.
    """

    if argv is None:
        argv = sys.argv

    # setup command line parser
    parser = E.ArgumentParser(description=__doc__)

    parser.add_argument("--task", dest="task", type=str,
                        choices=["extract_table", "get_coverage",
                                 "clean_table"],
                        help="task to perform")

    parser.add_argument("-t", "--table-name", dest="table", type=str,
                        help="table in SQLite DB to extract")

    # add common options (-h/--help, ...) and parse command line
    (args) = E.start(parser, argv=argv, add_database_options=True)

    if args.task == "extract_table":
        out_df = getTableFromDb(args.database_url, args.table)

    elif args.task == "get_coverage":
        out_df = getModelCoverage(args.database_url,
                                  table_regex="(\S+)_transcript_counts")

    elif args.task == "clean_table":
        infile = argv[-1]
        out_df = cleanStatsTable(infile)

    out_df.to_csv(args.stdout,
                  sep="\t", index_label="track")

    # write footer and output benchmark information.
    E.stop()


if __name__ == "__main__":
    sys.exit(main(sys.argv))



'''gff2table.py - compute features for intersection of two gff files
=================================================================

:Tags: Genomics Intervals Annotation Comparison GFF

Purpose
-------

collect intervals from two gff files and compute features based on
their intersection. The script is intended to compute properties for a
set of non-overlapping windows.

Transforms:
   * none:        no transform
   * overlap:     overlap between set1 and set2
   * complement:  part of set1 that is not covered by set2
   * third_codon: only takes every third position. Needs frame information
                  in the gff file.

Decorators:
   * GC:            G+C content of intervals
   * count:         number of windows
   * mean-length:   mean length of intervals overlapping with window

Usage
-----

Example::

   python gff2table.py --help

Type::

   python gff2table.py --help

for command line help.

Command line options
--------------------

'''
import sys
import re

import cgatcore.experiment as E
import cgat.Genomics as Genomics
import cgat.IndexedFasta as IndexedFasta
import cgat.IndexedGenome as IndexedGenome

import cgat.Intervals as Intervals
import cgat.Stats as Stats
import cgat.GTF as GTF


def decorator_counts(intervals, start, end, contig, fasta):
    """compute length distribution."""
    d = Stats.DistributionalParameters([x[1] - x[0] for x in intervals])
    return d['nval'], str(d)


def decorator_percent_coverage(intervals, start, end, contig, fasta):
    """compute length of intervals."""
    d = Stats.DistributionalParameters([x[1] - x[0] for x in intervals])
    return 100.0 * float(d['sum']) / (end - start), str(d)


def decorator_mean_length(intervals, start, end, contig, fasta):
    """compute length distribution."""
    d = Stats.DistributionalParameters([x[1] - x[0] for x in intervals])
    return d['mean'], str(d)


def decorator_median_length(intervals, start, end, contig, fasta):
    """compute length distribution."""
    d = Stats.DistributionalParameters([x[1] - x[0] for x in intervals])
    return d['median'], str(d)


def decorator_percent_gc(intervals, start, end, contig, fasta):
    """compute G+C content in intervals.
    """
    l, ngc = 0, 0

    # fetch sequence of the complete window first
    sequence = fasta.getSequence(contig, "+", start, end)

    for istart, iend in intervals:
        ngc += len([x for x in sequence[istart - start:iend - start] if x in "GCgc"])
        l += iend - istart

    return 100.0 * ngc / l, None


def decorator_median_score(values, start, end, contig):
    """compute median of values."""
    d = Stats.DistributionalParameters(values)
    return d['median'], str(d)


def decorator_mean_score(values, start, end, contig):
    """compute mean of values."""
    d = Stats.DistributionalParameters(values)
    return d['mean'], str(d)


def decorator_stddev_score(values, start, end, contig):
    """compute stddev of values."""
    d = Stats.DistributionalParameters(values)
    return d['stddev'], str(d)


def decorator_min_score(values, start, end, contig):
    """compute minumum of values."""
    d = Stats.DistributionalParameters(values)
    return d['min'], str(d)


def decorator_max_score(values, start, end, contig):
    """compute minumum of values."""
    d = Stats.DistributionalParameters(values)
    return d['max'], str(d)


def transform_overlap(start, end, intervals_with_gff):
    """transform: overlap of intervals in x with y."""
    y = Intervals.combineIntervals(
        [(x[0], x[1]) for x in intervals_with_gff])
    return Intervals.pruneIntervals(y, start, end)


def transform_complement(start, end, intervals_with_gff):
    y = Intervals.combineIntervals(
        [(x[0], x[1]) for x in intervals_with_gff])
    return Intervals.complementIntervals(y, start, end)


def transform_third_codon(start, end, intervals_with_gff):
    """transform: only return nucleotide positions in window (start, end)
    that are in third codon position.
    """
    intervals = []
    for istart, iend, gff in intervals_with_gff:

        if gff.frame == ".":
            raise ValueError("need a frame for third codon positions.")

        # frame = nucleotides from start to next codon
        frame = int(gff.frame)

        # to make life easier, convert to 0-based coordinates,
        # with zero starting at first position in window
        # re-arrange positions on negative strand
        if Genomics.IsNegativeStrand(gff.strand):
            # convert to negative strand coordinates counting from 0
            coordinate_offset = end
            reverse = True
            istart, iend = end - iend, end - istart
        else:
            istart, iend = istart - start, iend - start
            reverse = False
            coordinate_offset = start

        # make sure that you start on a second codon position and within window
        if istart < 0:
            frame = (frame + istart) % 3
            istart = 0
        if frame != 0:
            istart -= (3 - frame)
        istart += 2

        iend = min(iend, end - start)

        for x in range(istart, iend, 3):

            if reverse:
                c = coordinate_offset - x - 1
            else:
                c = coordinate_offset + x
            intervals.append((c, c + 1))

    return Intervals.combineIntervals(intervals)


def test_transform_third_codon():

    def test_entry(frame, strand, xfrom, xto, start, end, ref):

        entry = GTF.Entry()
        entry.frame = frame
        entry.strand = strand
        entry.start = xfrom
        entry.end = xto

        intervals = transform_third_codon(start, end, [(xfrom, xto, entry)])
        if ref != intervals:
            print("failed:", ref != intervals)

    test_entry(0, "+", 1, 7, 0, 6, [(3, 4)])
    test_entry(0, "-", 1, 7, 0, 6, [(1, 2), (4, 5)])
    test_entry(1, "+", 1, 7, 0, 6, [(1, 2), (4, 5)])
    test_entry(2, "+", 1, 7, 0, 6, [(2, 3), (5, 6)])
    test_entry(1, "-", 1, 7, 0, 6, [(3, 4)])
    test_entry(2, "-", 1, 7, 0, 6, [(2, 3), (5, 6)])

    sys.exit(0)


def annotateWindows(contig, windows, gff_data, fasta, options):
    """annotate windows."""

    index = IndexedGenome.IndexedGenome()
    for g in gff_data:
        index.add(g.contig, g.start, g.end, g)

    is_gtf = options.is_gtf

    if options.transform == "none":
        transform = lambda x, y, z: [(x[0], x[1]) for x in z]
    elif options.transform == "overlap":
        transform = transform_overlap
    elif options.transform == "complement":
        transform = transform_complement
    elif options.transform == "third_codon":
        transform = transform_third_codon
    else:
        raise ValueError("unknown transform %s" % options.transform)

    work_on_intervals = True
    if options.decorator == "counts":
        decorator = decorator_counts
    elif options.decorator == "mean-length":
        decorator = decorator_mean_length
    elif options.decorator == "median-length":
        decorator = decorator_median_length
    elif options.decorator == "percent-coverage":
        decorator = decorator_percent_coverage
    elif options.decorator == "gc":
        decorator = decorator_percent_gc
    elif options.decorator == "median-score":
        decorator = decorator_median_score
        work_on_intervals = False
    elif options.decorator == "mean-score":
        decorator = decorator_mean_score
        work_on_intervals = False
    elif options.decorator == "stddev-score":
        decorator = decorator_stddev_score
        work_on_intervals = False
    elif options.decorator == "min-score":
        decorator = decorator_min_score
        work_on_intervals = False
    elif options.decorator == "max-score":
        decorator = decorator_max_score
        work_on_intervals = False
    else:
        raise ValueError("unknown decorator %s" % options.decorator)

    for start, end in windows:

        # counts/length before/after transformation
        n1, l1, n2, l2 = 0, 0, 0, 0

        values, intervals_with_gff, genes, transcripts = [], [], set(), set()

        try:
            for istart, iend, value in index.get(contig, start, end):
                n1 += 1
                l1 += iend - istart
                intervals_with_gff.append((istart, iend, value))
                values.append(value.score)
                if is_gtf:
                    genes.add(value.gene_id)
                    transcripts.add(value.transcript_id)
        except KeyError:
            pass

        if n1 == 0 and options.skip_empty:
            continue

        if work_on_intervals:

            if options.loglevel >= 3:
                options.stdlog.write("# intervals in window %i:%i before transformation: %s\n" % (
                    start, end, str(intervals)))

            intervals = transform(start, end, intervals_with_gff)

            for xstart, xend in intervals:
                n2 += 1
                l2 += xend - xstart

            if options.loglevel >= 3:
                options.stdlog.write("# intervals in window %i:%i after transformation: %s\n" % (
                    start, end, str(intervals)))

            score, extra_info = decorator(intervals, start, end, contig, fasta)

        else:
            if len(values) > 0:
                values = list(map(float, values))
                score, extra_info = decorator(values, start, end, contig)
            else:
                score, extra_info = 0, None

            l2 = 0
            n2 = 0

        if is_gtf:
            ngenes, ntranscripts = len(genes), len(transcripts)
        else:
            ngenes, ntranscripts = 0, 0

        if extra_info:
            extra_info = re.sub("\t", ";", extra_info)
        options.stdout.write("\t".join(
            map(str, (contig, start, end,
                      ngenes, ntranscripts,
                      n1, l1,
                      n2, l2,
                      score,
                      extra_info))) + "\n")


def main(argv=None):
    """script main.

    parses command line options in sys.argv, unless *argv* is given.
    """

    if argv is None:
        argv = sys.argv

    parser = E.ArgumentParser(description=__doc__)

    parser.add_argument("--version", action='version', version="1.0")

    parser.add_argument(
        "-g", "--genome-file", dest="genome_file", type=str,
        help="filename with genome (indexed).")

    parser.add_argument(
        "-w", "--windows-bed-file", dest="filename_windows", type=str,
        help="gff file with windows to use.")

    parser.add_argument(
        "-d", "--filename-data", dest="filename_data", type=str,
        help="gff file with data to use.")

    parser.add_argument("--is-gtf", dest="is_gtf", action="store_true",
                        help="filename-data is gtf file")

    parser.add_argument(
        "-f", "--features", dest="features", type=str, action="append",
        choices=("GC", ),
        help="features to compute.")

    parser.add_argument(
        "-c", "--decorator", dest="decorator", type=str,
        choices=("counts", "gc", "gc3", "mean-length", "median-length",
                 "percent-coverage",
                 "median-score", "mean-score", "stddev-score", "min-score",
                 "max-score"),
        help="decorators to use.")

    parser.add_argument(
        "-e", "--skip-empty", dest="skip_empty", action="store_true",
        help="skip empty windows.")

    parser.add_argument(
        "-t", "--transform=", dest="transform", type=str,
        choices=(
            "none", "overlap", "complement", "third_codon"),
        help="transform to use when mapping overlapping regions onto window.")

    parser.set_defaults(
        genome_file=None,
        filename_windows=None,
        filename_data=None,
        features=[],
        skip_empty=False,
        decorator="counts",
        transform="none",
        is_gtf=False,
    )

    (args) = E.start(parser)

    #    test_transform_third_codon()

    if not args.filename_windows:
        raise ValueError("please supply a gff file with window information.")

    if args.loglevel >= 1:
        args.stdlog.write("# reading windows...")
        args.stdlog.flush()

    windows = GTF.readAsIntervals(
        GTF.iterator(iotools.open_file(args.filename_windows, "r")))

    if args.loglevel >= 1:
        args.stdlog.write("done\n")
        args.stdlog.flush()

    if args.filename_data:
        if args.loglevel >= 1:
            args.stdlog.write("# reading data...")
            args.stdlog.flush()

        if args.is_gtf:
            gff_data = GTF.readFromFile(
                iotools.open_file(args.filename_data, "r"))
        else:
            gff_data = GTF.readFromFile(
                IOTOols.open_file(args.filename_data, "r"))

        if args.loglevel >= 1:
            args.stdlog.write("done\n")
            args.stdlog.flush()

        data_ranges = GTF.SortPerContig(gff_data)
    else:
        # use windows to compute properties
        # by supplying no data and asking for the complement = original window
        gff_data = None
        data_ranges = None
        args.transform = "complement"

    map_contig2size = {}

    if args.genome_file:
        fasta = IndexedFasta.IndexedFasta(args.genome_file)
        map_contig2size = fasta.getContigSizes()
    else:
        for contig, values in list(windows.items()):
            map_contig2size[contig] = max(lambda x: x[1], values)
        fasta = None

    contigs = list(map_contig2size.keys())
    contigs.sort()

    # proceed contig wise
    noutput_contigs, ncontigs_skipped_windows, ncontigs_skipped_data = 0, 0, 0

    args.stdout.write("\t".join(
        map(str, ("contig", "start", "end",
                  "ngenes", "ntranscripts",
                  "n1", "l1",
                  "n2", "l2",
                  "score",
                  "extra_info"))) + "\n")

    for contig in contigs:

        skip = False
        if contig not in windows:
            ncontigs_skipped_windows += 1
            skip = True

        if data_ranges and contig not in data_ranges:
            ncontigs_skipped_data += 1
            skip = True

        if skip:
            continue

        noutput_contigs += 1
        if data_ranges:
            annotateWindows(contig,
                            windows[contig],
                            gff_data[
                                data_ranges[contig][0]:data_ranges[contig][1]],
                            fasta,
                            args)
        else:
            annotateWindows(contig,
                            windows[contig],
                            [],
                            fasta,
                            args)

    E.info("ninput_windows=%i, noutput_contigs=%i, ninput_contigs=%i, nskipped_windows=%i, nskipped_data=%i" %
           (len(windows), noutput_contigs, len(contigs), ncontigs_skipped_windows, ncontigs_skipped_data))

    E.stop()


if __name__ == "__main__":
    sys.exit(main(sys.argv))



'''
data2histogram.py - histogram data in a table
=============================================

:Tags: Python

Purpose
-------

This script computes histograms over one or more
columns of a table.

Usage
-----

Example::

   python data2histogram.py --help

Type::

   python data2histogram.py --help

for command line help.

Command line options
--------------------

'''
import sys
import cgatcore.experiment as E
import cgat.Histogram as Histogram
import numpy


def main(argv=None):

    if not argv:
        argv = sys.argv

    parser = E.ArgumentParser(description=__doc__)

    parser.add_argument("--version", action='version', version="1.0")

    parser.add_argument("-r", "--range", dest="range", type=str,
                        help="range to calculate histogram for.")
    parser.add_argument("-b", "--bin-size", dest="bin_size", type=str,
                        help="bin size.")
    parser.add_argument("-i", "--titles", dest="titles", action="store_true",
                        help="use supplied column titles.")
    parser.add_argument("--no-null", dest="nonull", action="store_true",
                        help="do not output null values")
    parser.add_argument("--no-titles", dest="titles", action="store_false",
                        help="no column titles given.")
    parser.add_argument("-c", "--columns", dest="columns", type=str,
                        help="columns to take for calculating histograms.")
    parser.add_argument("--min-data", dest="min_data", type=int,
                        help="minimum amount of data required, if less data, then the histogram will be empty [default=%default].")
    parser.add_argument("--min-value", dest="min_value", type=float,
                        help="minimum value for histogram.")
    parser.add_argument("--max-value", dest="max_value", type=float,
                        help="maximum value for histogram.")
    parser.add_argument("--no-empty-bins", dest="no_empty_bins", action="store_true",
                        help="do not display empty bins.")
    parser.add_argument("--with-empty-bins", dest="no_empty_bins", action="store_false",
                        help="display empty bins.")
    parser.add_argument("--normalize", dest="normalize", action="store_true",
                        help="normalize histogram.")
    parser.add_argument("--cumulative", dest="cumulative", action="store_true",
                        help="calculate cumulative histogram.")
    parser.add_argument("--reverse-cumulative", dest="reverse_cumulative", action="store_true",
                        help="calculate reverse cumulative histogram.")
    parser.add_argument("--header-names", dest="headers", type=str,
                        help="use the following headers.")
    parser.add_argument("--ignore-out-of-range", dest="ignore_out_of_range", action="store_true",
                        help="ignore values that are out of range (as opposed to truncating them to range border.")
    parser.add_argument("--missing-value", dest="missing_value", type=str,
                        help="entry for missing values .")
    parser.add_argument("--use-dynamic-bins", dest="dynamic_bins", action="store_true",
                        help="each value constitutes its own bin.")
    parser.add_argument("--on-the-fly", dest="on_the_fly", action="store_true",
                        help="on the fly computation of histograms. Requires setting of min-value, max-value and bin_size.")

    parser.set_defaults(
        bin_size=None,
        range=None,
        titles=True,
        columns="all",
        append=(),
        no_empty_bins=True,
        min_value=None,
        max_value=None,
        normalize=False,
        cumulative=False,
        reverse_cumulative=False,
        nonull=None,
        ignore_out_of_range=False,
        min_data=1,
        headers=None,
        missing_value="na",
        dynamic_bins=False,
        on_the_fly=False,
        bin_format="%.2f",
        value_format="%6.4f",
    )

    (args) = E.start(parser)

    if args.columns != "all":
        args.columns = [int(x) - 1 for x in args.columns.split(",")]

    if args.range:
        args.min_value, args.max_value = list(map(
            float, args.range.split(",")))

    if args.headers:
        args.headers = args.headers.split(",")

    if args.on_the_fly:
        if args.min_value is None or args.max_value is None or \
           args.bin_size is None:
            raise ValueError("please supply columns, min-value, max-value and "
                             "bin-size for on-the-fly computation.")

        # try to glean titles from table:
        if args.titles:
            while 1:
                line = sys.stdin.readline()
                if not line:
                    break
                if line[0] == "#":
                    continue
                data = line[:-1].split("\t")
                break

            if args.columns == "all":
                args.titles = data
                args.columns = list(range(len(data)))
            else:
                args.titles = [data[x] for x in args.columns]

        bins = numpy.arange(
            args.min_value, args.max_value, float(args.bin_size))
        hh = Histogram.fillHistograms(
            sys.stdin, args.columns, [bins for x in range(len(args.columns))])
        n = len(hh)

        titles = ['bin']

        if args.headers:
            titles.append(args.headers[x])
        elif args.titles:
            titles.append(args.titles[x])
        else:
            for x in args.columns:
                titles.append("col%i" % (x + 1))

        if len(titles) > 1:
            args.stdout.write("\t".join(titles) + "\n")

        for x in range(len(bins)):
            v = []
            v.append(args.bin_format % bins[x])
            for c in range(n):
                v.append(args.value_format % hh[c][x])

            args.stdout.write("\t".join(v) + "\n")

    else:
        # in-situ computation of histograms
        # retrieve data
        first = True
        vals = []

        # parse data, convert to floats
        for l in args.stdin:

            if l[0] == "#":
                continue

            data = l[:-1].split("\t")

            if first:
                first = False
                ncols = len(data)
                if args.columns == "all":
                    args.columns = list(range(ncols))

                vals = [[] for x in args.columns]

                if args.titles:
                    try:
                        args.titles = [data[x] for x in args.columns]
                    except IndexError:
                        raise IndexError("not all columns %s found in data %s" % (
                            str(args.columns), str(data)))
                    continue

            for x in range(len(args.columns)):

                try:
                    v = float(data[args.columns[x]])
                except IndexError:
                    print("# IndexError in line:", l[:-1])
                    continue
                except ValueError:
                    continue

                vals[x].append(v)

        lines = None

        hists = []
        titles = []

        if not vals:
            if args.loglevel >= 1:
                args.stdlog.write("# no data\n")
            E.stop()
            sys.exit(0)

        for x in range(len(args.columns)):

            if args.loglevel >= 1:
                args.stdlog.write(
                    "# column=%i, num_values=%i\n" % (args.columns[x], len(vals[x])))

            if len(vals[x]) < args.min_data:
                continue

            h = Histogram.Calculate(vals[x],
                                    no_empty_bins=args.no_empty_bins,
                                    increment=args.bin_size,
                                    min_value=args.min_value,
                                    max_value=args.max_value,
                                    dynamic_bins=args.dynamic_bins,
                                    ignore_out_of_range=args.ignore_out_of_range)

            if args.normalize:
                h = Histogram.Normalize(h)
            if args.cumulative:
                h = Histogram.Cumulate(h)
            if args.reverse_cumulative:
                h = Histogram.Cumulate(h, direction=0)

            hists.append(h)

            for m in args.append:
                if m == "normalize":
                    hists.append(Histogram.Normalize(h))

            if args.headers:
                titles.append(args.headers[x])
            elif args.titles:
                titles.append(args.titles[x])
            else:
                titles.append("col%i" % args.columns[x])

        if titles:
            args.stdout.write("bin\t" + "\t".join(titles) + "\n")

        if len(hists) == 1:
            Histogram.Print(hists[0], nonull=args.nonull,
                            format_bin=args.bin_format)
        else:
            combined_histogram = Histogram.Combine(
                hists, missing_value=args.missing_value)
            Histogram.Print(combined_histogram,
                            nonull=args.nonull,
                            format_bin=args.bin_format)

    E.stop()


if __name__ == "__main__":
    sys.exit(main(sys.argv))



"""Compare phasing information in two VCF files.

"""

import sys
import os
import pysam
import collections
import string
import cgatcore.experiment as E


def main(argv=sys.argv):

    parser = E.ArgumentParser(description=__doc__)

    parser.add_argument("--version", action='version', version="1.0")

    parser.add_argument(
        "-i", "--input-vcf", dest="input_vcf_file", type=str,
        help="input vcf file")

    parser.add_argument(
        "-t", "--truth-vcf", dest="truth_vcf_file", type=str,
        help="truth vcf file")

    parser.add_argument(
        "-f", "--input-fasta", dest="input_fasta_file", type=str,
        help="input fasta file. faidx indexed reference sequence file to "
        "determine INDEL context ")

    parser.add_argument(
        "-e", "--input-bed", dest="input_bed_file", type=str,
        help="input file with intervals. Tab-delimited file of intervals "
        "in bed format to restrict analysis to. ")

    parser.add_argument(
        "-m", "--method", dest="methods", action="append", type=str,
        choices=("mutational-signature", "kinship"),
        help="methods to apply ")

    parser.set_defaults(
        methods=[],
        input_vcf_file=None,
        input_bed_file=None,
        input_fasta_file=None,
        truth_vcf_file=None,
    )

    (args, unknown) = E.start(parser,
                              argv,
                              add_output_options=True,
                              unknowns=True)

    if len(unknown) == 1:
        args.input_vcf_file = unknown[0]

    if args.input_vcf_file is None:
        raise ValueError("please supply a VCF file")

    if args.truth_vcf_file is None:
        raise ValueError("please supply a VCF file with truth data")

    if args.input_fasta_file is None:
        raise ValueError("please supply a fasta file with the reference genome")

    if not os.path.exists(args.input_vcf_file):
        raise OSError("input vcf file {} does not exist".format(
            args.input_vcf_file))

    if not os.path.exists(args.input_vcf_file + ".tbi"):
        raise OSError("input vcf file {} needs to be indexed".format(
            args.input_vcf_file))

    if not os.path.exists(args.truth_vcf_file):
        raise OSError("truth vcf file {} does not exist".format(
            args.truth_vcf_file))

    if not os.path.exists(args.truth_vcf_file + ".tbi"):
        raise OSError("truth vcf file {} needs to be indexed".format(
            args.truth_vcf_file))

    if not os.path.exists(args.input_fasta_file):
        raise OSError("input fasta file {} does not exist".format(
            args.input_fasta_file))

    if not os.path.exists(args.input_fasta_file + ".fai"):
        raise OSError("input fasta file {} needs to be indexed".format(
            args.input_fasta_file))

    # update paths to absolute
    args.input_fasta_file = os.path.abspath(args.input_fasta_file)
    args.input_vcf_file = os.path.abspath(args.input_vcf_file)
    args.truth_vcf_file = os.path.abspath(args.truth_vcf_file)

    test_vcf = pysam.VariantFile(args.input_vcf_file)
    truth_vcf = pysam.VariantFile(args.truth_vcf_file)
    contigs = test_vcf.header.contigs
    truth_contigs = set(truth_vcf.header.contigs)

    test_vcf_samples = set(test_vcf.header.samples)
    truth_vcf_samples = set(truth_vcf.header.samples)

    common_samples = test_vcf_samples.intersection(truth_vcf_samples)
    if len(common_samples) == 0:
        raise ValueError("no common samples in test/truth VCFs")

    def pair_iterator(test_vcf, truth_vcf, contig):
        counter = E.Counter()
        test_iter = test_vcf.fetch(contig)
        truth_iter = truth_vcf.fetch(contig)

        test_record = next(test_iter)
        truth_record = next(truth_iter)
        try:
            while 1:
                if test_record.pos < truth_record.pos:
                    test_record = next(test_iter)
                    continue

                elif test_record.pos > truth_record.pos:
                    truth_record = next(truth_iter)
                    continue

                elif len(test_record.alts) > 1:
                    counter.skip_test_truth += 1
                    test_record = next(test_iter)
                    continue

                elif len(truth_record.alts) > 1:
                    counter.skip_multiallelic_truth += 1
                    truth_record = next(truth_iter)
                    continue

                elif test_record.alts != truth_record.alts:
                    counter.skip_genotype_difference += 1
                    test_record = next(test_iter)
                    truth_record = next(truth_iter)
                    continue

                if test_record.ref != truth_record.ref:
                    # todo: deal with indels
                    raise ValueError(
                        "mismatching reference bases at position "
                        "{}:{}".format(test_record.chrom, test_record.pos))

                yield test_record, truth_record
                test_record = next(test_iter)
                truth_record = next(truth_iter)

        except StopIteration:
            pass

        E.debug(str(counter))

    counters_per_contig = {}

    for contig in contigs:
        counter_contig = collections.defaultdict(E.Counter)
        counters_per_contig[contig] = counter_contig

        E.info("processing contig {}".format(contig))

        if contig not in truth_contigs:
            E.warn(
                "skipping contig {} as it is not in truth data".format(contig))
            continue

        switch = False
        last_is_unphased = True

        for test_record, truth_record in pair_iterator(test_vcf, truth_vcf, contig):

            for sample in common_samples:
                counter = counter_contig[sample]

                truth_phased = truth_record.samples[sample].phased
                test_phased = test_record.samples[sample].phased
                truth_genotype = truth_record.samples[sample]["GT"]
                test_genotype = test_record.samples[sample]["GT"]
                truth_alleles = set(truth_genotype)
                test_alleles = set(test_genotype)

                ignore = False
                if not truth_phased:
                    counter.truth_unphased += 1
                    ignore = True
                if not test_phased:
                    counter.test_unphased += 1
                    ignore = True
                    last_is_unphased = True
                else:
                    last_is_unphased = False

                if len(test_alleles) == 1:
                    counter.test_homozygous += 1
                    ignore = True
                else:
                    if not test_phased:
                        counter.test_unphased_hets += 1

                if len(truth_alleles) == 1:
                    counter.truth_homozygous += 1
                    ignore = True

                if ignore:
                    counter.ignore += 1
                    continue

                E.debug("comparing: {}:{} {} -> {}: {} {}".format(
                    test_record.chrom, test_record.pos,
                    test_record.ref, test_record.alts,
                    test_genotype,
                    truth_genotype))

                if switch:
                    truth_genotype = truth_genotype[::-1]

                counter.test_phased_hets += 1

                if truth_genotype != test_genotype:
                    if not last_is_unphased:
                        E.debug("SWITCH: {}".format(switch))
                        counter.switch += 1
                    switch = not switch

    outf = args.stdout
    outf.write("\t".join(("contig",
                          "sample",
                          "switch_error_percent",
                          "false_negative_rate",
                          "switches",
                          "test_phased_hets",
                          "test_unphased_hets",
                          "test_unphased",
                          "truth_unphased",
                          "test_homozygous",
                          "truth_homozygous")) + "\n")

    for contig, contig_dict in list(counters_per_contig.items()):
        for sample, c in list(contig_dict.items()):
            outf.write("\t".join(
                map(str, (
                    contig,
                    sample,
                    "{:6.4f}".format(100.0 * c.switch / (c.test_phased_hets + 1)),
                    "{:6.4f}".format(100.0 * c.test_unphased_hets /
                                     (c.test_phased_hets + c.test_unphased_hets)),
                    c.switch,
                    c.test_phased_hets,
                    c.test_unphased_hets,
                    c.test_unphased,
                    c.truth_unphased,
                    c.test_homozygous,
                    c.truth_homozygous))) + "\n")

    E.stop()
    # TODO: aggregate stats per samples



'''combine_tables.py - join tables
==================================

:Tags: Python

Purpose
-------

This script reads several tab-separated tables and joins them into a
single one.

Usage
-----

The option ``--header-names`` sets the column titles explicitely. Add
``--skip-titles`` if you want to avoid echoing the original title in
the input files.


Example::

   python combine_tables.py --help

Type::

   python combine_tables.py --help

for command line help.

Command line options
--------------------

'''
import sys
import re
import os
import glob
import collections
import pandas

import cgatcore.iotools as iotools
import cgatcore.experiment as E


def read_tables(filenames, *args, **kwargs):

    tables = []
    for filename in filenames:
        try:
            table = pandas.read_csv(filename, *args, **kwargs)
        except pandas.errors.EmptyDataError:
            E.warn("file '{}' is empty".format(filename))
            continue
        except pandas.errors.ParserError as ex:
            E.warn("file '{}' has parsing error: {}".format(filename, ex))
            continue
        if len(table) == 0:
            E.warn("table '{}' is empty".format(filename))
            continue
        tables.append((table, filename))
    return zip(*tables)


def concatenate_tables(filenames,
                       regex_filename="(\S+)",
                       separator="\t",
                       headers=None,
                       missing_value=None,
                       cat=None):

    '''concatenate tables.'''

    rx = re.compile(regex_filename)

    tables, filenames = read_tables(filenames, sep=separator)

    if headers is None or headers == "auto":
        row_headers = [
            [y for y in rx.search(x).groups()] for x in filenames]
    else:
        row_headers = [headers]

    if cat is None:
        if len(row_headers) == 1:
            row_head_titles = ["filename"]
        else:
            row_head_titles = ["pattern" + str(x) for x in range(len(row_headers))]
    else:
        row_head_titles = [x.strip() for x in cat.split(",")]
        if len(row_headers[0]) != len(row_head_titles):
            raise ValueError(
                "row header (%i) has different number of fields in "
                "regular expression than supplied by the --cat option (%i)" %
                (len(row_headers[0]), len(row_head_titles)))

    # avoid MultiIndex if only single level
    if len(row_head_titles) == 1:
        row_head_titles = row_head_titles
        row_headers = [x[0] for x in row_headers]

    df = pandas.concat(tables, axis=0, keys=row_headers, names=row_head_titles) \
               .reset_index() \
               .drop(["level_1"], axis=1)
    return df


def main(argv=sys.argv):

    parser = E.ArgumentParser(description=__doc__)

    parser.add_argument("--version", action='version', version="1.0")

    parser.add_argument("-t", "--no-titles",
                        dest="input_has_titles",
                        action="store_false",
                        help="no titles in input .")

    parser.add_argument("--ignore-titles",
                        dest="ignore_titles",
                        action="store_true",
                        help="ignore titles in input ")

    parser.add_argument("-i", "--skip-titles",
                        dest="skip_titles",
                        action="store_true",
                        help="skip output of titles.")

    parser.add_argument("-m", "--missing-value",
                        dest="missing_value",
                        type=str,
                        help="entry to use for missing values.")

    parser.add_argument("--header-names", dest="headers", type=str,
                        help="add headers for files as a ,-separated "
                        "list .")

    parser.add_argument("-c", "--columns", dest="columns", type=str,
                        help="columns to use for joining. Multiple columns "
                        "can be specified as a comma-separated list ")

    parser.add_argument("-k", "--take",
                        dest="take",
                        type=str,
                        action="append",
                        help="columns to take. If not set, all columns "
                        "except for "
                        "the join columns are taken ")

    parser.add_argument("-g", "--glob", dest="glob", type=str,
                        help="wildcard expression for table names.")

    parser.add_argument(
        "-s", "--sort-order", dest="sort", type=str,
        help="sort by column titles in particular given order: "
        "alphabetical|numeric|list of columns.")

    parser.add_argument(
        "-e", "--merge-overlapping", dest="merge", action="store_true",
        help="simply merge tables without matching up "
        "rows.")

    parser.add_argument("-a", "--cat", dest="cat", type=str,
                        help="simply concatenate tables. Adds an "
                        "additional column called X with the filename ")

    parser.add_argument("--sort-keys", dest="sort_keys", type=str,
                        choices=("numeric", "alphabetic"),
                        help="sort key columns by value.")

    parser.add_argument("--keep-empty", dest="ignore_empty",
                        action="store_false",
                        help="keep empty tables. The default is "
                        "to ignore them.")

    parser.add_argument("--ignore-empty",
                        dest="ignore_empty",
                        action="store_true",
                        help="ignore empty tables - this is "
                        "the default .")

    parser.add_argument("--add-file-prefix",
                        dest="add_file_prefix",
                        action="store_true",
                        help="add file prefix to "
                        "columns headers. Suitable for multi-column"
                        "tables")

    parser.add_argument("--use-file-prefix",
                        dest="use_file_prefix",
                        action="store_true",
                        help="use file prefix as column headers. "
                        "Suitable for two-column tables ")

    parser.add_argument("--prefixes", dest="prefixes", type=str,
                        help="list of prefixes to use. "
                        ", separated list of prefixes. "
                        "The number of prefixes need to correspond to the "
                        "number of input files")

    parser.add_argument("--regex-filename", dest="regex_filename",
                        type=str,
                        help="pattern to apply to filename to "
                        "build prefix")

    parser.add_argument("--regex-start",
                        dest="regex_start",
                        type=str,
                        help="regular expression to start "
                        "collecting table in a file")

    parser.add_argument("--regex-end",
                        dest="regex_end",
                        type=str,
                        help="regular expression to end collecting "
                        "table in a file")

    parser.add_argument("--sep",
                        dest="separator",
                        type=str,
                        help="table separator to use. The default is to use tabs. ")

    parser.add_argument("--test", dest="test",
                        type=int,
                        help="test combining tables with "
                        "first X rows")

    parser.set_defaults(
        input_has_titles=True,
        skip_titles=False,
        missing_value=None,
        headers=None,
        sort=None,
        glob=None,
        columns="1",
        sort_keys=False,
        merge=False,
        ignore_empty=True,
        regex_start=None,
        regex_end=None,
        add_file_prefix=False,
        use_file_prefix=False,
        cat=None,
        take=[],
        regex_filename="(.*)",
        prefixes=None,
        test=0,
        separator="\t",
    )

    (args, unknown) = E.start(parser,
                              argv=argv,
                              unknowns=True)

    if args.headers:
        if "," in args.headers:
            args.headers = args.headers.split(",")
        else:
            args.headers = re.split("\s+", args.headers.strip())

    if args.sort and args.sort not in ("numeric", "alphabetic"):
        if "," in args.sort:
            args.sort = args.sort.split(",")
        else:
            args.sort = re.split("\s+", args.sort)

    if args.merge:
        args.columns = []
    else:
        args.columns = [int(x) - 1 for x in args.columns.split(",")]

    args.filenames = []

    if args.glob:
        args.filenames += glob.glob(args.glob)

    args.filenames += unknown

    if len(args.filenames) < 1:
        raise ValueError("no tables found.")

    E.info("combining %i tables" % len(args.filenames))

    if args.cat:
        table = concatenate_tables(args.filenames,
                                   regex_filename=args.regex_filename,
                                   separator=args.separator,
                                   headers=args.headers,
                                   missing_value=args.missing_value,
                                   cat=args.cat)

    table.to_csv(args.stdout, sep=args.separator, index=False)
    E.stop()


if __name__ == '__main__':
    sys.exit(main(sys.argv))



'''cgat2dot.py - create a graph between cgat scripts
====================================================

:Tags: Python

Purpose
-------

This script creates an rdf description of a cgat script.

Optionally, the script outputs also a galaxy xml description of the
scripts' interface.

Usage
-----

Example::

   python cgat2dot.py scripts/*.py

Type::

   python cgat2dot.py --help

for command line help.

Documentation
-------------

Command line options
--------------------

'''

import os
import sys
import re
import imp

import cgatcore.experiment as E

BASE_URL = "https://www.cgat.org/downloads/public/cgat/documentation/"

ORIGINAL_START = None

PARSER = None


def _e(string):
    return string.replace(' ', '_')


MAP_FORMATS = {
    'tsv': 'table',
    'table': 'table',
    'stats': 'table',
    'csv': 'table',
}

PRINCIPAL_FORMATS = ('bam',
                     'gff',
                     'gtf',
                     'bed',
                     'wiggle',
                     'fasta',
                     'fastq',
                     'fastqs')

BREAK_FORMATS = {'table': 0}
MAP_TYPE2FORMAT = {
    'gff': 'gff,gtf',
    'gtf': 'gff,gtf',
    'bam': 'bam',
    'sam': 'sam',
    'bigwig': 'bigWig',
    'bed': 'bed',
}

NODE_STYLE_DEFAULT = 'color="#A5BB00",style="filled"'
NODE_STYLE_FORMAT = 'color="#7577B8",style="filled"'

EDGE_STYLE_CONVERSION = 'color="#7577B8",penwidth=2'
EDGE_STYLE_DEFAULT = 'color="#A5BB00",penwidth=1'


class DummyError(Exception):
    pass


def LocalStart(parser, *args, **kwargs):
    '''stub for E.start - set return_parser argument to true'''
    global PARSER
    PARSER = ORIGINAL_START(parser,
                            return_parser=True,
                            **kwargs
                            )
    raise DummyError()


def getDescription(scriptname, docstring):
    '''get script description from docstring.'''

    description = scriptname
    for line in docstring.split("\n"):
        if line.startswith(scriptname):
            description = line[line.index("-") + 1:].strip()
            break

    return description


def guessFormats(scriptname, docstring):
    '''guess the input/output format of a script.'''

    input_format, output_format = "tsv", "tsv"

    if "2" in scriptname:
        input_format, output_format = scriptname.split("2")

    # map cgat format names to GALAXY ones
    input_format = MAP_FORMATS.get(input_format, input_format)
    output_format = MAP_FORMATS.get(output_format, output_format)

    return input_format, output_format


def buildParam(**kwargs):
    '''return a parameter with default values.

    Specific fields can be set by providing keyword arguments.
    '''

    param = {}

    param['label'] = "label"
    param['description'] = "description"
    param['rank'] = 1
    param['display'] = 'show'
    param['min_occurrence'] = 0
    param['max_occurrence'] = 1

    # get default value
    param['value'] = "value"
    param['type'] = "text"
    param['dependencies'] = {}
    param['property_bag'] = {}
    param['arg_long'] = '--long-argument'

    param.update(kwargs)
    return param


def processScript(script_name, outfile, args):
    '''process one script.'''

    # call other script
    prefix, suffix = os.path.splitext(script_name)

    dirname = os.path.dirname(script_name)
    basename = os.path.basename(script_name)[:-3]

    if args.src_dir:
        dirname = args.src_dir
        script_name = os.path.join(dirname, basename) + ".py"

    if os.path.exists(prefix + ".pyc"):
        os.remove(prefix + ".pyc")

    pyxfile = os.path.join(dirname, "_") + basename + ".pyx"
    if os.path.exists(pyxfile):
        pass

    try:
        module = imp.load_source(basename, script_name)
    except ImportError as msg:
        E.warn('could not import %s - skipped: %s' % (basename, msg))
        return

    E.info("loaded module %s" % module)

    E.start = LocalStart
    try:
        module.main(argv=["--help"])
    except TypeError as msg:
        E.warn('could not import %s: %s' % (basename, msg))
        return
    except DummyError:
        pass

    # get script's docstring
    docstring = module.__doc__

    input_format, output_format = guessFormats(basename, docstring)

    if output_format in BREAK_FORMATS:
        nodename = '%s%i' % (output_format, BREAK_FORMATS[output_format])
        outfile.write('%s [label="%s"];\n' %
                      (nodename,
                       output_format))
        BREAK_FORMATS[output_format] += 1
        output_format = nodename

    url = BASE_URL + "scripts/%s.html" % basename

    # Note that URL needs to be uppercase!
    if input_format in PRINCIPAL_FORMATS and \
       output_format in PRINCIPAL_FORMATS:
        edge_style = EDGE_STYLE_CONVERSION
    else:
        edge_style = EDGE_STYLE_DEFAULT
    outfile.write('"%s" -> "%s" [label="%s",URL="%s",%s];\n' %
                  (input_format, output_format, basename, url,
                   edge_style))

    return

    # for k in dir(PARSER):
    #     print k, getattr(PARSER, k)
    # for option in PARSER.option_list:
    # print option, option.type, option.help, option._short_opts,
    # option._long_opts, option.default

    # @prefix clp: <http://www.humgen.nl/climate/ontologies/clp#> .
    # @prefix co: <http://www.isi.edu/ikcap/Wingse/componentOntology.owl#> .
    # @prefix dcterms: <http://purl.org/dc/terms/> .

    defaults = PARSER.get_default_values()

    for option in PARSER.option_list:
        # ignore options added by optparse
        if option.dest is None:
            continue

        # ignore benchmarking options
        if option.dest.startswith("timeit"):
            continue

        # ignore options related to forcing output
        if "force" in option.dest:
            continue

        # ignore some special options:
        # if option.dest in ("output_filename_pattern", ):
        #    continue

        # ignore output options
        if option.dest in ("stdin", "stdout", "stdlog", "stderr", "loglevel"):
            continue

        # remove default from help string
        option.help = re.sub("\[[^\]]*%default[^\]]*\]", "", option.help)

        param = buildParam()

        # get command line option call (long/short option)
        try:
            param['arg'] = option._short_opts[0]
        except IndexError:
            pass

        try:
            param['arg_long'] = option._long_opts[0]
        except IndexError:
            pass

        assert 'arg' in param or 'arg_long' in param

        # print "----------------------------------"
        # print [(x,getattr(option,x)) for x in dir( option )]

        param['name'] = option.dest
        param['ns_name'] = option.dest
        if option.type == "int":
            param['type'] = "integer"
        elif option.type == "float":
            param['type'] = "float"
        elif option.type == "string":
            param['type'] = "text"
            if option.metavar:
                mvar = option.metavar.lower()
                if mvar in MAP_TYPE2FORMAT:
                    param['format'] = MAP_TYPE2FORMAT[mvar]
                    param['type'] = "data"
                if mvar == "bam":
                    pass

        elif option.type == "choice":
            param['type'] = "select"
            param['choices'] = option.choices
            if option.action == "append":
                param['multiple'] = True
        elif option.action.startswith("store"):
            param['type'] = "boolean"
        else:
            raise ValueError("unknown type for %s" % str(option))

        param['label'] = option.dest
        param['description'] = option.help
        param['rank'] = 1
        param['display'] = 'show'
        param['min_occurrence'] = 0
        param['max_occurrence'] = 1

        # get default value
        param['value'] = getattr(defaults,  option.dest)


def main(argv=None):
    """script main.

    parses command line options in sys.argv, unless *argv* is given.
    """

    if not argv:
        argv = sys.argv

    # setup command line parser
    parser = E.ArgumentParser(description=__doc__)

    parser.add_argument("-f", "--format", dest="output_format", type=str,
                        choices=("rdf", "galaxy"),
                        help="output format . ")

    parser.add_argument("-l", "--list", dest="filename_list", type=str,
                        help="filename with list of files to export "
                        ". ")

    parser.add_argument("-s", "--source-dir", dest="src_dir", type=str,
                        help="directory to look for scripts . ")

    parser.add_argument("-r", "--input-regex", dest="input_regex", type=str,
                        help="regular expression to extract script name "
                        ". ")

    parser.add_argument("-p", "--output-filename-pattern", dest="output_pattern",
                        type=str,
                        help="pattern to build output filename. Should contain "
                        "an '%s' . ")

    parser.set_defaults(output_format="rdf",
                        src_dir=None,
                        input_regex=None,
                        output_pattern=None,
                        filename_list=None)

    # add common options (-h/--help, ...) and parse command line
    (args) = E.start(parser, argv=argv)

    if len(args) == 0:
        E.info("reading script names from stdin")
        for line in args.stdin:
            if line.startswith("#"):
                continue
            args.append(line[:-1].split("\t")[0])

    # start script in order to build the command line parser
    global ORIGINAL_START
    ORIGINAL_START = E.start

    if args.output_pattern and not args.input_regex:
        raise ValueError(
            "please specify --input-regex when using --output-filename-pattern")

    outfile = args.stdout
    outfile.write("""digraph cgat {
    size="10,20";
    # scale graph so that there are no overlaps
    overlap=scale;
    splines=True;
\n""")

    # set node format for principal genomic formats
    for format in PRINCIPAL_FORMATS:
        outfile.write('"%s" [shape=box,%s];\n' % (format, NODE_STYLE_FORMAT))

    # general node format
    outfile.write('node [%s];\n' % NODE_STYLE_DEFAULT)

    # go through script to provide edges
    for script_name in args:
        if not script_name.endswith(".py"):
            raise ValueError("expected a python script ending in '.py'")

        E.info("input=%s, output=%s" % (script_name, outfile))
        processScript(script_name, outfile, args)

    outfile.write("}\n")

    E.stop()


if __name__ == "__main__":
    sys.exit(main(sys.argv))



'''bams2bam.py - merge genomic and transcriptome mapped bamfiles
====================================================================

:Tags: Genomics NGS Geneset BAM Manipulation

Purpose
-------

This script takes as input two BAM files from an RNASeq experiment.
The first bam file (:file:`bamG`) should contain reads mapped against
the genome using a mapper permitting splicing (e.g. tophat). The
second bam file (:file:`bamT`) should contain reads mapped against
known transcripts. This script will write a new bam file that removes
reads from :term:`bamG` that map to regions that are conflicting with
those in :term:`bamT`.

.. note::
   Note that if junctions are supplied, the resultant bam files will not
   be sorted by position.

.. glossary::

   bamG
      :term:`bam` formatted file with reads mapped against the genome

   bamT
      :term:`bam` formatted file with reads mapped against transcripts

Usage
-----

Example::

   python bams2bam.py bamT.bam bamG.bam

Type::

   python bams2bam.py --help

for command line help.

Documentation
-------------

The script needs to look-up reads via their names. It thus builds an
index of reads mapping

This script requires the NM attributes to be set. If it is not set,
you will need to set a policy.

Command line options
--------------------

'''

import os
import sys
import pysam

import cgatcore.experiment as E
import cgat.GTF as GTF
import cgatcore.iotools as iotools
import cgat.Bed as Bed
import cgat.IndexedGenome as IndexedGenome
from cgat.BamTools.bamtools import bams2bam_filter


def main(argv=None):
    """script main.

    parses command line options in sys.argv, unless *argv* is given.
    """

    if not argv:
        argv = sys.argv

    # setup command line parser
    parser = E.ArgumentParser(version="%prog version: $Id$",
                              usage=globals()["__doc__"])

    parser.add_argument(
        "-g", "--gtf-file", dest="filename_gtf", type=str,
        help="filename with gene models in gtf format ")

    parser.add_argument(
        "-m", "--filename-mismapped", dest="filename_mismapped", type=str,
        help="output bam file for mismapped reads ")

    parser.add_argument(
        "-j", "--junctions-bed-file", dest="filename_junctions", type=str,
        help="bam file with reads mapped across junctions ")

    parser.add_argument(
        "-r", "--filename-regions", dest="filename_regions", type=str,
        help="filename with regions to remove in bed format ")

    parser.add_argument(
        "-t", "--transcripts-gtf-file", dest="filename_transcriptome",
        type=str,
        help="bam file with reads mapped against transcripts ")

    parser.add_argument(
        "-p", "--map-tsv-file", dest="filename_map", type=str,
        help="filename mapping transcript numbers (used by "
        "--filename-transciptome) to transcript names "
        "(used by --filename-gtf) ")

    parser.add_argument(
        "-s", "--filename-stats", dest="filename_stats", type=str,
        help="filename to output stats to ")

    parser.add_argument(
        "-o", "--colour",
        dest="colour_mismatches", action="store_true",
        help="mismatches will use colour differences (CM tag) ")

    parser.add_argument(
        "-i", "--ignore-mismatches",
        dest="ignore_mismatches", action="store_true",
        help="ignore mismatches ")

    parser.add_argument(
        "-c", "--remove-contigs", dest="remove_contigs", type=str,
        help="','-separated list of contigs to remove ")

    parser.add_argument(
        "-f", "--force-output", dest="force", action="store_true",
        help="force overwriting of existing files ")

    parser.add_argument("-u", "--unique", dest="unique", action="store_true",
                        help="remove reads not matching uniquely ")

    parser.add_argument("--output-sam", dest="output_sam", action="store_true",
                        help="output in sam format ")

    parser.set_defaults(
        filename_gtf=None,
        filename_mismapped=None,
        filename_junctions=None,
        filename_transcriptome=None,
        filename_map=None,
        remove_contigs=None,
        force=False,
        unique=False,
        colour_mismatches=False,
        ignore_mismatches=False,
        output_sam=False,
        filename_table=None,
    )

    # add common options (-h/--help, ...) and parse command line
    (options, args) = E.start(parser, argv=argv)

    if len(args) != 1:
        raise ValueError("please supply one bam file")

    bamfile_genome = args[0]
    genome_samfile = pysam.AlignmentFile(bamfile_genome, "rb")

    if options.remove_contigs:
        options.remove_contigs = options.remove_contigs.split(",")

    if options.filename_map:
        E.info("reading map")
        id_map = iotools.read_map(
            iotools.open_file(options.filename_map), has_header=True)
        id_map = dict([(y, x) for x, y in id_map.items()])
    else:
        id_map = None

    transcripts = {}
    if options.filename_gtf:
        E.info("indexing geneset")
        mapped, missed = 0, 0
        for gtf in GTF.transcript_iterator(
                GTF.iterator(iotools.open_file(options.filename_gtf))):
            gtf.sort(key=lambda x: x.start)
            transcript_id = gtf[0].transcript_id
            if id_map:
                try:
                    transcript_id = id_map[transcript_id]
                    mapped += 1
                except KeyError:
                    missed += 1
                    continue
            transcripts[transcript_id] = gtf

        E.info("read %i transcripts from geneset (%i mapped, %i missed)" %
               (len(transcripts), mapped, missed))

    regions_to_remove = None
    if options.filename_regions:
        E.info("indexing regions")
        regions_to_remove = IndexedGenome.Simple()
        for bed in Bed.iterator(iotools.open_file(options.filename_regions)):
            regions_to_remove.add(bed.contig, bed.start, bed.end)
        E.info("read %i regions" % len(regions_to_remove))

    if options.filename_transcriptome:
        transcripts_samfile = pysam.AlignmentFile(options.filename_transcriptome,
                                                  "rb")
    else:
        transcripts_samfile = None

    if options.output_sam:
        output_samfile = pysam.AlignmentFile("-", "wh", template=genome_samfile)
    else:
        output_samfile = pysam.AlignmentFile("-", "wb", template=genome_samfile)

    if options.filename_mismapped:
        if not options.force and os.path.exists(options.filename_mismapped):
            raise IOError("output file %s already exists" %
                          options.filename_mismapped)
        output_mismapped = pysam.AlignmentFile(options.filename_mismapped,
                                               "wb",
                                               template=genome_samfile)
    else:
        output_mismapped = None

    if options.filename_junctions:
        junctions_samfile = pysam.AlignmentFile(options.filename_junctions,
                                                "rb")
    else:
        junctions_samfile = None

    c = bams2bam_filter(genome_samfile,
                        output_samfile,
                        output_mismapped,
                        transcripts_samfile,
                        junctions_samfile,
                        transcripts,
                        regions=regions_to_remove,
                        unique=options.unique,
                        remove_contigs=options.remove_contigs,
                        colour_mismatches=options.colour_mismatches,
                        ignore_mismatches=options.ignore_mismatches,
                        ignore_transcripts=transcripts_samfile is None,
                        ignore_junctions=junctions_samfile is None)

    if options.filename_stats:
        outf = iotools.open_file(options.filename_stats, "w")
        outf.write("category\tcounts\n%s\n" % c.asTable())
        outf.close()

    if options.filename_transcriptome:
        transcripts_samfile.close()

    genome_samfile.close()
    output_samfile.close()
    if output_mismapped:
        output_mismapped.close()

    # write footer and output benchmark information.
    E.stop()


if __name__ == "__main__":
    sys.exit(main(sys.argv))
